# Natural Language Toolkit: API for alignment and translation objects
#
# Copyright (C) 2001-2026 NLTK Project
# Author: Will Zhang <wilzzha@gmail.com>
#         Guan Gui <ggui@student.unimelb.edu.au>
#         Steven Bird <stevenbird1@gmail.com>
#         Tah Wei Hoon <hoon.tw@gmail.com>
# URL: <https://www.nltk.org/>
# For license information, see LICENSE.TXT

import subprocess
from collections import namedtuple


class AlignedSent:
    """
    Return an aligned sentence object, which encapsulates two sentences
    along with an ``Alignment`` between them.

    Typically used in machine translation to represent a sentence and
    its translation.

        >>> from nltk.translate import AlignedSent, Alignment
        >>> algnsent = AlignedSent(['klein', 'ist', 'das', 'Haus'],
        ...     ['the', 'house', 'is', 'small'], Alignment.fromstring('0-3 1-2 2-0 3-1'))
        >>> algnsent.words
        ['klein', 'ist', 'das', 'Haus']
        >>> algnsent.mots
        ['the', 'house', 'is', 'small']
        >>> algnsent.alignment
        Alignment([(0, 3), (1, 2), (2, 0), (3, 1)])
        >>> from nltk.corpus import comtrans
        >>> print(comtrans.aligned_sents()[54])
        <AlignedSent: 'Weshalb also sollten...' -> 'So why should EU arm...'>
        >>> print(comtrans.aligned_sents()[54].alignment)
        0-0 0-1 1-0 2-2 3-4 3-5 4-7 5-8 6-3 7-9 8-9 9-10 9-11 10-12 11-6 12-6 13-13

    :param words: Words in the target language sentence
    :type words: list(str)
    :param mots: Words in the source language sentence
    :type mots: list(str)
    :param alignment: Word-level alignments between ``words`` and ``mots``.
        Each alignment is represented as a 2-tuple (words_index, mots_index).
    :type alignment: Alignment
    """

    def __init__(self, words, mots, alignment=None):
        self._words = words
        self._mots = mots
        if alignment is None:
            self.alignment = Alignment([])
        else:
            assert type(alignment) is Alignment
            self.alignment = alignment

    @property
    def words(self):
        return self._words

    @property
    def mots(self):
        return self._mots

    def _get_alignment(self):
        return self._alignment

    def _set_alignment(self, alignment):
        _check_alignment(len(self.words), len(self.mots), alignment)
        self._alignment = alignment

    alignment = property(_get_alignment, _set_alignment)

    def __repr__(self):
        """
        Return a string representation for this ``AlignedSent``.

        :rtype: str
        """
        words = "[%s]" % (", ".join("'%s'" % w for w in self._words))
        mots = "[%s]" % (", ".join("'%s'" % w for w in self._mots))

        return f"AlignedSent({words}, {mots}, {self._alignment!r})"

    def _to_dot(self):
        """
        Dot representation of the aligned sentence
        """
        s = "graph align {\n"
        s += "node[shape=plaintext]\n"

        # Declare node
        s += "".join([f'"{w}_source" [label="{w}"] \n' for w in self._words])
        s += "".join([f'"{w}_target" [label="{w}"] \n' for w in self._mots])

        # Alignment
        s += "".join(
            [
                f'"{self._words[u]}_source" -- "{self._mots[v]}_target" \n'
                for u, v in self._alignment
            ]
        )

        # Connect the source words
        for i in range(len(self._words) - 1):
            s += '"{}_source" -- "{}_source" [style=invis]\n'.format(
                self._words[i],
                self._words[i + 1],
            )

        # Connect the target words
        for i in range(len(self._mots) - 1):
            s += '"{}_target" -- "{}_target" [style=invis]\n'.format(
                self._mots[i],
                self._mots[i + 1],
            )

        # Put it in the same rank
        s += "{rank = same; %s}\n" % (" ".join('"%s_source"' % w for w in self._words))
        s += "{rank = same; %s}\n" % (" ".join('"%s_target"' % w for w in self._mots))

        s += "}"

        return s

    def _repr_svg_(self):
        """
        Ipython magic : show SVG representation of this ``AlignedSent``.
        """
        dot_string = self._to_dot().encode("utf8")
        output_format = "svg"
        try:
            process = subprocess.Popen(
                ["dot", "-T%s" % output_format],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except OSError as e:
            raise Exception("Cannot find the dot binary from Graphviz package") from e
        out, err = process.communicate(dot_string)

        return out.decode("utf8")

    def __str__(self):
        """
        Return a human-readable string representation for this ``AlignedSent``.

        :rtype: str
        """
        source = " ".join(self._words)[:20] + "..."
        target = " ".join(self._mots)[:20] + "..."
        return f"<AlignedSent: '{source}' -> '{target}'>"

    def invert(self):
        """
        Return the aligned sentence pair, reversing the directionality

        :rtype: AlignedSent
        """
        return AlignedSent(self._mots, self._words, self._alignment.invert())


class Alignment(frozenset):
    """
    A storage class for representing alignment between two sequences, s1, s2.
    In general, an alignment is a set of tuples of the form (i, j, ...)
    representing an alignment between the i-th element of s1 and the
    j-th element of s2.  Tuples are extensible (they might contain
    additional data, such as a boolean to indicate sure vs possible alignments).

        >>> from nltk.translate import Alignment
        >>> a = Alignment([(0, 0), (0, 1), (1, 2), (2, 2)])
        >>> a.invert()
        Alignment([(0, 0), (1, 0), (2, 1), (2, 2)])
        >>> print(a.invert())
        0-0 1-0 2-1 2-2
        >>> a[0]
        [(0, 1), (0, 0)]
        >>> a.invert()[2]
        [(2, 1), (2, 2)]
        >>> b = Alignment([(0, 0), (0, 1)])
        >>> b.issubset(a)
        True
        >>> c = Alignment.fromstring('0-0 0-1')
        >>> b == c
        True
    """

    def __new__(cls, pairs):
        self = frozenset.__new__(cls, pairs)
        self._len = max(p[0] for p in self) if self != frozenset([]) else 0
        self._index = None
        return self

    @classmethod
    def fromstring(cls, s):
        """
        Read a giza-formatted string and return an Alignment object.

            >>> Alignment.fromstring('0-0 2-1 9-2 21-3 10-4 7-5')
            Alignment([(0, 0), (2, 1), (7, 5), (9, 2), (10, 4), (21, 3)])

        :type s: str
        :param s: the positional alignments in giza format
        :rtype: Alignment
        :return: An Alignment object corresponding to the string representation ``s``.
        """

        return Alignment([_giza2pair(a) for a in s.split()])

    def __getitem__(self, key):
        """
        Look up the alignments that map from a given (left) index.
        Returns an empty list for an index that has no alignments.

        Only integer indices are supported. Slicing and other non-integer
        keys are rejected with ``TypeError`` rather than silently returning
        ``[]`` -- the sparse index has no contiguous range to slice, so a
        mistaken lookup should fail loudly instead of masking the bug.
        """
        if not isinstance(key, int):
            raise TypeError(
                "Alignment indices must be integers, not %s" % type(key).__name__
            )
        if self._index is None:
            self._build_index()
        return self._index.get(key, [])

    def invert(self):
        """
        Return an Alignment object, being the inverted mapping.
        """
        return Alignment(((p[1], p[0]) + p[2:]) for p in self)

    def range(self, positions=None):
        """
        Work out the range of the mapping from the given positions.
        If no positions are specified, compute the range of the entire mapping.
        """
        image = set()
        if self._index is None:
            self._build_index()
        if not positions:
            positions = self._index.keys()
        for p in positions:
            image.update(f for _, f in self._index.get(p, []))
        return sorted(image)

    def __repr__(self):
        """
        Produce a Giza-formatted string representing the alignment.
        """
        return "Alignment(%r)" % sorted(self)

    def __str__(self):
        """
        Produce a Giza-formatted string representing the alignment.
        """
        return " ".join("%d-%d" % p[:2] for p in sorted(self))

    def _build_index(self):
        """
        Build a sparse index mapping each left index ``i`` to the list of
        alignments originating from word ``i``.

        The index is keyed only by the left indices that actually occur, so its
        size is bounded by the number of pairs rather than by the largest left
        index (which is attacker-controlled in giza-format input). A dense
        ``[[] for _ in range(self._len + 1)]`` list would instead let a single
        tiny pair with a huge left index (e.g. ``"0-0 100000000-1"``) allocate
        ~100M empty lists -- gigabytes of memory -- an unbounded-allocation DoS
        (CWE-770; CVE-2026-12837).
        """
        self._index = {}
        for p in self:
            self._index.setdefault(p[0], []).append(p)


def _giza2pair(pair_string):
    i, j = pair_string.split("-")
    return int(i), int(j)


def _naacl2pair(pair_string):
    i, j, p = pair_string.split("-")
    return int(i), int(j)


def _check_alignment(num_words, num_mots, alignment):
    """
    Check whether the alignments are legal.

    :param num_words: the number of source language words
    :type num_words: int
    :param num_mots: the number of target language words
    :type num_mots: int
    :param alignment: alignment to be checked
    :type alignment: Alignment
    :raise IndexError: if alignment falls outside the sentence
    """

    assert type(alignment) is Alignment

    if not all(0 <= pair[0] < num_words for pair in alignment):
        raise IndexError("Alignment is outside boundary of words")
    if not all(pair[1] is None or 0 <= pair[1] < num_mots for pair in alignment):
        raise IndexError("Alignment is outside boundary of mots")


PhraseTableEntry = namedtuple("PhraseTableEntry", ["trg_phrase", "log_prob"])


class PhraseTable:
    """
    In-memory store of translations for a given phrase, and the log
    probability of the those translations
    """

    def __init__(self):
        self.src_phrases = dict()

    def translations_for(self, src_phrase):
        """
        Get the translations for a source language phrase

        :param src_phrase: Source language phrase of interest
        :type src_phrase: tuple(str)

        :return: A list of target language phrases that are translations
            of ``src_phrase``, ordered in decreasing order of
            likelihood. Each list element is a tuple of the target
            phrase and its log probability.
        :rtype: list(PhraseTableEntry)
        """
        return self.src_phrases[src_phrase]

    def add(self, src_phrase, trg_phrase, log_prob):
        """
        :type src_phrase: tuple(str)
        :type trg_phrase: tuple(str)

        :param log_prob: Log probability that given ``src_phrase``,
            ``trg_phrase`` is its translation
        :type log_prob: float
        """
        entry = PhraseTableEntry(trg_phrase=trg_phrase, log_prob=log_prob)
        if src_phrase not in self.src_phrases:
            self.src_phrases[src_phrase] = []
        self.src_phrases[src_phrase].append(entry)
        self.src_phrases[src_phrase].sort(key=lambda e: e.log_prob, reverse=True)

    def __contains__(self, src_phrase):
        return src_phrase in self.src_phrases



import importlib.util
import os
import subprocess
import sys
import tempfile
from pathlib import Path
from unittest import mock

import pytest


def test_module_hijacking_prevention():
    """Ensure imports of vulnerable modules from CWD are blocked."""
    parent_paths = [p for p in sys.path if p and p != "."]

    with tempfile.TemporaryDirectory() as d:
        with open(os.path.join(d, "joblib.py"), "w") as f:
            f.write("print('HIJACK_SUCCESS')\n")
        with open(os.path.join(d, "victim.py"), "w") as f:
            f.write(
                f"import sys\n"
                f"sys.path = {repr(parent_paths)} + sys.path\n"
                "from nltk.util import parallelize_preprocess\n"
                "list(parallelize_preprocess(str.upper, ['a'], processes=1))\n"
            )
        env = os.environ.copy()
        res = subprocess.run(
            [sys.executable, "victim.py"],
            cwd=d,
            env=env,
            capture_output=True,
            text=True,
        )
        assert "HIJACK_SUCCESS" not in res.stdout


def test_host_imports_of_vulnerable_modules_are_blocked():
    """CWD imports pulled in during `import nltk` are blocked."""
    parent_paths = [p for p in sys.path if p and p != "."]

    with tempfile.TemporaryDirectory() as d:
        with open(os.path.join(d, "joblib.py"), "w") as f:
            f.write("print('HOST_HIJACK')\n")
        with open(os.path.join(d, "victim.py"), "w") as f:
            f.write(
                f"import sys\n"
                f"sys.path = {repr(parent_paths)} + sys.path\n"
                "import nltk\n"
                "import joblib\n"
                "print('HOST_SUCCESS')\n"
            )
        env = os.environ.copy()
        res = subprocess.run(
            [sys.executable, "victim.py"],
            cwd=d,
            env=env,
            capture_output=True,
            text=True,
        )
        assert "HOST_HIJACK" not in res.stdout


def test_host_imports_of_non_vulnerable_modules_are_unaffected():
    """Host imports of non‑vulnerable modules from CWD succeed."""
    parent_paths = [p for p in sys.path if p and p != "."]

    with tempfile.TemporaryDirectory() as d:
        with open(os.path.join(d, "antigravity.py"), "w") as f:
            f.write("print('ANTIGRAVITY_HIJACK')\n")
        with open(os.path.join(d, "victim.py"), "w") as f:
            f.write(
                f"import sys\n"
                f"sys.path = [''] + {repr(parent_paths)} + sys.path\n"
                "import nltk\n"
                "import antigravity\n"
                "print('HOST_SUCCESS')\n"
            )
        env = os.environ.copy()
        env.pop("PYTHONSAFEPATH", None)  # this test needs the CWD searchable
        res = subprocess.run(
            [sys.executable, "victim.py"],
            cwd=d,
            env=env,
            capture_output=True,
            text=True,
        )
        try:
            assert res.returncode == 0
            assert "ANTIGRAVITY_HIJACK" in res.stdout
            assert "HOST_SUCCESS" in res.stdout
        except AssertionError:
            print("--- STDOUT ---\n", res.stdout)
            print("--- STDERR ---\n", res.stderr)
            raise


def test_disable_flag():
    """Ensure setting NLTK_DISABLE_IMPORT_SECURITY=1 disables the hook."""
    parent_paths = [p for p in sys.path if p and p != "."]

    with tempfile.TemporaryDirectory() as d:
        with open(os.path.join(d, "joblib.py"), "w") as f:
            f.write("print('DISABLED_HIJACK')\n")
        with open(os.path.join(d, "victim.py"), "w") as f:
            f.write(
                f"import os\n"
                f"os.environ['NLTK_DISABLE_IMPORT_SECURITY'] = '1'\n"
                f"import sys\n"
                f"sys.path = [''] + {repr(parent_paths)} + sys.path\n"
                "import nltk\n"
                "import joblib\n"
                "print('SUCCESS')\n"
            )
        env = os.environ.copy()
        env.pop("PYTHONSAFEPATH", None)  # must be able to reach the CWD module
        res = subprocess.run(
            [sys.executable, "victim.py"],
            cwd=d,
            env=env,
            capture_output=True,
            text=True,
        )
        try:
            assert (
                "DISABLED_HIJACK" in res.stdout
            ), "Malicious module should be loaded when hook is disabled"
        except AssertionError:
            print("--- STDOUT ---\n", res.stdout)
            print("--- STDERR ---\n", res.stderr)
            raise


@pytest.mark.skipif(
    not importlib.util.find_spec("joblib"), reason="joblib not installed"
)
def test_legitimate_import_from_site_packages():
    import joblib

    assert joblib.__file__ is not None


def test_pythonsafepath_is_propagated_for_child_interpreters():
    """
    Installing the hook must set PYTHONSAFEPATH=1 so freshly started worker
    interpreters (spawn/forkserver) inherit interpreter-level CWD isolation.
    """
    from nltk import inisec

    with mock.patch.dict(os.environ, {}, clear=False):
        os.environ.pop("PYTHONSAFEPATH", None)
        inisec._install()
        assert os.environ.get("PYTHONSAFEPATH") == "1"


def test_pythonsafepath_does_not_override_host_choice():
    """
    NLTK must not override a PYTHONSAFEPATH value the host has already set.
    """
    from nltk import inisec

    with mock.patch.dict(os.environ, {"PYTHONSAFEPATH": "0"}, clear=False):
        inisec._install()
        assert os.environ.get("PYTHONSAFEPATH") == "0"


@pytest.mark.xfail(
    strict=True,
    reason=(
        "Known residual limitation: fork-based joblib/loky workers inherit the "
        "parent's already-fixed sys.path and run on a fresh stack, so neither "
        "PYTHONSAFEPATH nor caller detection applies. Closing this in-library "
        "would require globally mutating the parent's sys.path, which this "
        "design deliberately avoids. Remedy: launch with -P / PYTHONSAFEPATH. "
        "If this test starts passing, the limitation was closed and the xfail "
        "must be removed."
    ),
)
@pytest.mark.skipif(
    not importlib.util.find_spec("joblib"), reason="joblib not installed"
)
def test_worker_process_cwd_import_is_blocked():
    """
    Documents the fork-worker residual: a module dropped in the CWD is imported
    by a joblib/loky worker because the worker's fresh stack has no NLTK frame
    and its inherited sys.path still contains the CWD. Expected to xfail under
    the current in-library-only design.
    """
    parent_paths = [p for p in sys.path if p and p != "."]

    with tempfile.TemporaryDirectory() as d:
        with open(os.path.join(d, "joblib_victim.py"), "w") as f:
            f.write("print('WORKER_HIJACK')\n")
        with open(os.path.join(d, "victim.py"), "w") as f:
            f.write(
                f"import sys\n"
                f"sys.path = {repr(parent_paths)} + sys.path\n"
                "import nltk\n"
                "from joblib import Parallel, delayed\n"
                "\n"
                "def work(_):\n"
                "    try:\n"
                "        import joblib_victim\n"
                "        return 'IMPORTED'\n"
                "    except ImportError:\n"
                "        return 'BLOCKED'\n"
                "\n"
                "if __name__ == '__main__':\n"
                "    out = Parallel(n_jobs=2)(delayed(work)(i) for i in range(2))\n"
                "    print('RESULTS', out)\n"
            )
        env = os.environ.copy()
        # Deliberately do NOT set PYTHONSAFEPATH here: this test characterises
        # the residual that exists WITHOUT the recommended launch-time remedy.
        env.pop("PYTHONSAFEPATH", None)
        res = subprocess.run(
            [sys.executable, "victim.py"],
            cwd=d,
            env=env,
            capture_output=True,
            text=True,
        )
        assert "WORKER_HIJACK" not in res.stdout
        assert "IMPORTED" not in res.stdout


@pytest.mark.skipif(
    sys.version_info < (3, 11),
    reason="PYTHONSAFEPATH was added in Python 3.11; on earlier versions it is "
    "silently ignored, so this launch-time remedy does not apply.",
)
def test_worker_process_is_protected_with_pythonsafepath():
    """
    Confirms the recommended remedy works: with PYTHONSAFEPATH=1 in the launch
    environment, worker interpreters omit the CWD and the hijack is prevented.
    Requires Python 3.11+, where PYTHONSAFEPATH exists.
    """
    if not importlib.util.find_spec("joblib"):
        pytest.skip("joblib not installed")

    parent_paths = [p for p in sys.path if p and p != "."]

    with tempfile.TemporaryDirectory() as d:
        with open(os.path.join(d, "joblib_victim.py"), "w") as f:
            f.write("print('WORKER_HIJACK')\n")
        with open(os.path.join(d, "victim.py"), "w") as f:
            f.write(
                f"import sys\n"
                f"sys.path = {repr(parent_paths)} + sys.path\n"
                "import nltk\n"
                "from joblib import Parallel, delayed\n"
                "\n"
                "def work(_):\n"
                "    try:\n"
                "        import joblib_victim\n"
                "        return 'IMPORTED'\n"
                "    except ImportError:\n"
                "        return 'BLOCKED'\n"
                "\n"
                "if __name__ == '__main__':\n"
                "    out = Parallel(n_jobs=2)(delayed(work)(i) for i in range(2))\n"
                "    print('RESULTS', out)\n"
            )
        env = os.environ.copy()
        env["PYTHONSAFEPATH"] = "1"
        res = subprocess.run(
            [sys.executable, "victim.py"],
            cwd=d,
            env=env,
            capture_output=True,
            text=True,
        )
        assert "WORKER_HIJACK" not in res.stdout


def test_wordnet_app_reference_decode_rejects_wrong_types():
    """
    nltk.app.wordnet_app.Reference.decode() unpickles attacker-controlled,
    base64-encoded data straight from the wordnet browser's lookup_ URLs via
    RestrictedUnpickler. Every malformed input must surface as ValueError,
    since that's the only exception type the lookup_ route in do_GET catches.
    """
    import base64
    import pickle

    from nltk.app.wordnet_app import Reference

    good = base64.urlsafe_b64encode(pickle.dumps(("dog", {}), -1)).decode()
    ref = Reference.decode(good)
    assert ref.word == "dog"
    assert ref.synset_relations == {}

    bad_payloads = [
        (42, {}),
        ([[[1]]], {}),
        (None, {}),
        ("dog", []),
        ("dog", {"dog.n.01": ["not", "a", "set"]}),
        ("dog", {"dog.n.01": frozenset()}),
        ("dog", {42: {"hypernym"}}),
        42,
        ("dog",),
        ("dog", {}, "extra"),
    ]
    for bad_payload in bad_payloads:
        bad = base64.urlsafe_b64encode(pickle.dumps(bad_payload, -1)).decode()
        try:
            Reference.decode(bad)
            raise AssertionError(f"Reference.decode should reject {bad_payload!r}")
        except ValueError:
            pass

    for bad_string in ["not valid base64!!!", "", "====", "aGVsbG8="]:
        try:
            Reference.decode(bad_string)
            raise AssertionError(f"Reference.decode should reject {bad_string!r}")
        except ValueError:
            pass



# Natural Language Toolkit: TnT Tagger
#
# Copyright (C) 2001-2026 NLTK Project
# Author: John Winstead <https://github.com/jhnwnstd>
#         Sam Huston <sjh900@gmail.com>
#
# URL: <https://www.nltk.org/>
# For license information, see LICENSE.TXT

"""
Implementation of 'TnT - A Statistical Part of Speech Tagger'
by Thorsten Brants

https://aclanthology.org/A00-1031.pdf

Where Brants (2000) is silent, the implementation makes principled
choices flagged with ``Underspecified by Brants:`` in inline comments
so future readers can tell paper-derived behavior from local policy:

* ``_compute_lambda`` - even-split tie-breaking when multiple ``c_i``
  reach the max, and zeroed weights on degenerate training input.
* ``train`` - empty sentences contribute no EOS counts.
* ``_expand_states`` - strict ``>`` beam tie-break, so the
  first-encountered hypothesis wins. No clearly better policy is
  known; future work on deterministic hypothesis recombination
  could improve this.
"""

from math import log2

from nltk.probability import ConditionalFreqDist, FreqDist
from nltk.tag.api import TaggerI

# Used in place of log2(p) when p underflows; log2(1e-300) ~= -996.58
# is still well above the negative-inf result of log2(0).
_LOG_FLOOR_2 = log2(1e-300)

# Sentinel tags used at sentence boundaries.
_BOS = ("BOS", False)
_EOS = ("EOS", False)

_SENT_MARKS = (".", "!", "?", ";")

# Returned by transition-cache misses so inner-loop ``.get`` calls never
# need an extra ``None`` check before the next lookup.
_EMPTY_DICT: dict = {}


def _safe_log2(p):
    """``log2(p)`` for a non-zero probability, ``_LOG_FLOOR_2`` otherwise."""
    return log2(p) if p > 1e-300 else _LOG_FLOOR_2


def _safe_inverse(n):
    """``1/n`` for a non-zero ``n``, ``0.0`` otherwise so callers can use
    multiplication without an inline guard."""
    return (1.0 / n) if n else 0.0


class TnT(TaggerI):
    """
    TnT statistical part of speech tagger.

    This class implements the Trigrams'n'Tags tagger described by
    Brants (2000). TnT is a second order hidden Markov model. Its
    hidden states are part of speech tags, its observations are word
    tokens, and each transition is conditioned on the two previous tag
    states.

    Input

    TnT estimates and scores sentence boundary transitions, so training
    and tagging data should normally be sentence delimited.

    Training data should be a list of tagged sentences. Each sentence
    should be a list of ``(word, tag)`` tuples.

    ``tag()`` expects one tokenized sentence.
    ``tagdata()`` expects a list of tokenized sentences.

    For simple punctuation based segmentation of token streams, use
    ``basic_sent_chop()`` or pass ``segment=True`` to ``tag()``. The
    punctuation tokens are ``.``, ``!``, ``?``, and ``;``, following
    Brants's sentence boundary heuristic.

    Model

    For a sentence of length ``T``, the decoder chooses the tag sequence
    that maximizes

        prod_i P(t_i | t_{i-1}, t_{i-2}) * P(w_i | t_i)

    and then scores an explicit end of sentence transition. EOS is
    recorded as an ordinary next state in the trained n-gram stream,
    so its factor uses the same deleted interpolation as the body
    transitions (unigram, bigram, and trigram tiers). This generalizes
    Brants's printed factor ``P(t_{T+1} | t_T)``, which is bigram only.

    Decoding runs in log space over a Viterbi trellis whose state is
    the pair ``(t_{i-1}, t_i)``.

    Transition smoothing

    Transition probabilities use context independent deleted
    interpolation of unigram, bigram, and trigram tag models.

        P(t_i | t_{i-1}, t_{i-2})
            = l1 * P(t_i)
            + l2 * P(t_i | t_{i-1})
            + l3 * P(t_i | t_{i-1}, t_{i-2})

    The interpolation weights ``l1``, ``l2``, and ``l3`` are estimated
    from the training counts by the deleted interpolation procedure in
    Brants (2000), section 2.2.

    Emissions

    Known words use a tag dictionary. For a known word, candidate tags
    are restricted to the tags assigned to that word in the training
    data. The emission probability is the relative frequency of the
    word and tag pair in the training lexicon.

    Unknown words use Brants's suffix model by default. The model builds
    capitalization specific suffix tries from infrequent training words,
    using the threshold of at most 10 occurrences from Brants (2000),
    section 2.3. At decode time it uses the longest suffix observed in
    training, capped at 10 characters, smooths the suffix tag
    distribution by successive abstraction, and applies Bayesian
    inversion to obtain emission like scores.

    A user supplied tagger may be passed with ``unk``. When supplied,
    it overrides the built in suffix model for unknown words.

    Beam search

    The decoder uses beam pruning to limit memory and runtime. After
    each Viterbi step, states whose log probability is worse than the
    current best state by more than ``log2(N)`` are discarded. The
    default threshold ``N=1000`` follows Brants (2000), section 2.5.

    Capitalization

    When ``C=True``, capitalization is included in the tag state. This
    is equivalent to splitting each tag into capitalized and
    uncapitalized variants, as described by Brants (2000), section 2.4.

    Reproducibility

    Training-time iteration is sorted in ``_compute_lambda``,
    ``_build_suffix_model``, and ``_tagword``'s candidate construction,
    so the trained model and decoded output are bit-identical across
    reorderings of equivalent training data.
    """

    def __init__(self, unk=None, Trained=False, N=1000, C=False):
        """
        Construct a TnT statistical tagger. The tagger must be trained
        before it can be used to tag input.

        :param unk: instance of a POS tagger, conforms to TaggerI.
                    When supplied, overrides the built-in suffix model
                    on the unknown-word path.
        :type unk: TaggerI
        :param Trained: Indication that the POS tagger is trained or not.
                        Set True to skip training the optional ``unk``
                        tagger on the next train() call.
        :type Trained: bool
        :param N: Beam search pruning threshold. After each Viterbi
                step any state whose log-probability is worse than
                the best by more than a factor of N is discarded.
                Must be a positive integer; 1000 is a good default.
        :type N: int
        :param C: Capitalization flag. When True, tags are differentiated
                by whether the source word is capitalized. This rarely
                improves accuracy in practice.
        :type C: bool
        """

        # ``bool`` is an ``int`` subclass; the explicit bool check rejects
        # True/False, which would otherwise pass the int type check.
        if isinstance(N, bool) or not isinstance(N, int) or N < 1:
            raise ValueError(f"N must be a positive integer, got {N!r}")

        self._beam_threshold = N
        self._use_capitalization = C
        self._unk = unk
        self._unk_trained = Trained

        self._tag_unigrams = FreqDist()
        self._tag_bigrams = ConditionalFreqDist()
        self._tag_trigrams = ConditionalFreqDist()
        self._word_tag_freqs = ConditionalFreqDist()

        self._lambda1 = 0.0
        self._lambda2 = 0.0
        self._lambda3 = 0.0
        self._num_tag_tokens = 0
        self._log2_beam_threshold = 0.0

        # Unknown-word decoding uses a capitalization split suffix model, a
        # raw tag prior, and theta for successive abstraction smoothing.
        self._suffix_trie_by_cap = {
            False: ConditionalFreqDist(),
            True: ConditionalFreqDist(),
        }
        self._tag_prior_probs = {}
        self._theta = 0.0

        # Read by ``_expand_states`` instead of recomputing the deleted
        # interpolation and ``log2`` per beam expansion.
        self._trans_logp_unigram = {}
        self._trans_logp_bigram = {}
        self._trans_logp_trigram = {}

        # Cleared on every train() so contents always reflect current model state.
        self._candidate_tags_cache = {}

        self.unknown = 0
        self.known = 0

    def train(self, data):
        """
        Trains the tagger on a list of tagged sentences. Each call
        rebuilds the model from scratch on the supplied data. The
        n-gram counts, word-tag lexicon, suffix model, and
        deleted-interpolation weights are all replaced, and the decode
        cache is cleared.

        The optional external unknown-word tagger (``unk``) is trained
        on the supplied data only the first time ``train()`` is called.
        Subsequent calls leave it alone, since retraining it on each
        new training set is rarely what callers want.

        :param data: list of lists of (word, tag) tuples
        :type data: list[list[tuple[str, str]]]
        """

        # These structures accumulate corpus statistics, so retraining must
        # rebuild them from scratch rather than layer new counts on top.
        self._candidate_tags_cache.clear()
        self._tag_unigrams = FreqDist()
        self._tag_bigrams = ConditionalFreqDist()
        self._tag_trigrams = ConditionalFreqDist()
        self._word_tag_freqs = ConditionalFreqDist()

        unk = self._unk
        if unk is not None and not self._unk_trained:
            unk.train(data)

        word_tag_freqs = self._word_tag_freqs
        tag_unigrams = self._tag_unigrams
        tag_bigrams = self._tag_bigrams
        tag_trigrams = self._tag_trigrams
        cap_on = self._use_capitalization

        for sent in data:
            state_i_minus_2 = _BOS
            state_i_minus_1 = _BOS
            sent_has_tokens = False

            for word, tag in sent:
                sent_has_tokens = True
                c_i = cap_on and bool(word) and word[0].isupper()
                state_i = (tag, c_i)

                word_tag_freqs[word][tag] += 1
                tag_unigrams[state_i] += 1
                tag_bigrams[state_i_minus_1][state_i] += 1
                tag_trigrams[(state_i_minus_2, state_i_minus_1)][state_i] += 1

                state_i_minus_2, state_i_minus_1 = state_i_minus_1, state_i

            # Underspecified by Brants: EOS is treated as an ordinary next
            # state in the n-gram model, but empty sentences are skipped
            # so BOS does not acquire EOS as a spurious successor.
            if sent_has_tokens:
                tag_unigrams[_EOS] += 1
                tag_bigrams[state_i_minus_1][_EOS] += 1
                tag_trigrams[(state_i_minus_2, state_i_minus_1)][_EOS] += 1

        self._compute_lambda()

        # This total intentionally includes EOS because the unigram model
        # and deleted interpolation are estimated over the same event stream.
        self._num_tag_tokens = tag_unigrams.N()
        self._log2_beam_threshold = log2(self._beam_threshold)

        self._build_transition_logp_cache()
        self._build_suffix_model()

        self._unk_trained = True

    def tag(self, tokens, segment=False):
        """
        Tag a single sentence. Delegates the actual decode to
        `_tagword`, then pairs each chosen tag with its input token.

        When `segment` is True, the input may contain mid-sequence
        sentence punctuation [.!?;]. The decoder splits on those
        tokens and re-seeds the BOS state for each segment.
        The default is False because most NLTK callers pre-segment,
        and auto-splitting on `.` would mis-handle abbreviations
        like "Mr." in unsegmented input.

        :param tokens: words to tag
        :type tokens: list[str]
        :param segment: split on [.!?;] and decode each segment with a
                        fresh BOS state
        :type segment: bool
        :return: list of `(word, tag)` tuples
        """
        # Catch "forgot to tokenize" mistake before it produces garbage tags.
        if isinstance(tokens, (str, bytes)):
            raise TypeError(
                "tag() expects a list of tokens, not a single string or bytes. "
                "Tokenize first (e.g. tokens.split() or word_tokenize(s))."
            )
        if segment:
            return self._tag_segmented(tokens)
        if not (sent := list(tokens)):
            return []
        return self._pair_decoded(sent, self._tagword(sent))

    def tagdata(self, data, segment=False):
        """
        Tags a list of sentences. Each input sentence is a list of words;
        each output sentence is a list of (word, tag) tuples.

        :param data: list of list of words
        :type data: list[list[str]]
        :param segment: forwarded to ``tag``. Pass True to auto-split
                        each input on internal [.!?;] punctuation.
        :type segment: bool
        :return: list of list of (word, tag) tuples
        """
        if isinstance(data, (str, bytes)):
            raise TypeError(
                "tagdata() expects a list of tokenized sentences, "
                "not a single string or bytes object."
            )
        return [self.tag(sent, segment=segment) for sent in data]

    def _compute_lambda(self):
        """
        Computes the deleted-interpolation weights l1, l2, l3 from the
        observed tag n-grams. Tied maxima split the trigram count evenly
        among the winning lambdas. Branches with a zero denominator
        contribute zero.

        For each trigram (t1, t2, t3) with positive count we compare

            c1 = (f(t3) - 1) / (N - 1)
            c2 = (f(t2, t3) - 1) / (f(t2) - 1)
            c3 = (f(t1, t2, t3) - 1) / (f(t1, t2) - 1)
        """

        tag_unigrams = self._tag_unigrams
        tag_bigrams = self._tag_bigrams
        tag_trigrams = self._tag_trigrams
        unigram_n_minus_1 = tag_unigrams.N() - 1

        lambda1_mass = 0.0
        lambda2_mass = 0.0
        lambda3_mass = 0.0

        # Sorted iteration makes lambda mass accumulation order-independent.
        for state_i_minus_2, state_i_minus_1 in sorted(tag_trigrams.conditions()):
            trigram_dist = tag_trigrams[(state_i_minus_2, state_i_minus_1)]
            bigram_dist = tag_bigrams[state_i_minus_1]

            trigram_n_minus_1 = trigram_dist.N() - 1
            bigram_n_minus_1 = bigram_dist.N() - 1

            for state_i, count in sorted(trigram_dist.items()):
                # Subtracting one leaves the current event out, so each score
                # asks which model order would best predict this tag if this
                # occurrence were held out.
                c1 = (
                    (tag_unigrams[state_i] - 1) / unigram_n_minus_1
                    if unigram_n_minus_1
                    else 0.0
                )
                c2 = (
                    (bigram_dist[state_i] - 1) / bigram_n_minus_1
                    if bigram_n_minus_1
                    else 0.0
                )
                c3 = (count - 1) / trigram_n_minus_1 if trigram_n_minus_1 else 0.0

                # Underspecified by Brants: the trigram's count is credited
                # to the model order with the strongest held-out estimate.
                # Splitting ties evenly across the winning lambdas avoids
                # introducing an arbitrary preference between orders.
                maxc = max(c1, c2, c3)
                w1 = c1 == maxc
                w2 = c2 == maxc
                w3 = c3 == maxc
                share = count / (w1 + w2 + w3)

                if w1:
                    lambda1_mass += share
                if w2:
                    lambda2_mass += share
                if w3:
                    lambda3_mass += share

        # Underspecified by Brants: normalization turns the accumulated
        # winning mass into mixture weights. Zeroing on degenerate input
        # (no positive trigram mass) avoids a divide-by-zero and prevents
        # stale weights from a previous training run from leaking through.
        total_mass = lambda1_mass + lambda2_mass + lambda3_mass
        if total_mass > 0:
            self._lambda1 = lambda1_mass / total_mass
            self._lambda2 = lambda2_mass / total_mass
            self._lambda3 = lambda3_mass / total_mass
        else:
            self._lambda1 = 0.0
            self._lambda2 = 0.0
            self._lambda3 = 0.0

    def _build_transition_logp_cache(self):
        """
        Precompute transition log probabilities for Viterbi expansion.

        The decoder scores each transition with the same deleted
        interpolation used by the model.

            P(t_i | t_{i-2}, t_{i-1})
              = l1 P(t_i)
              + l2 P(t_i | t_{i-1})
              + l3 P(t_i | t_{i-2}, t_{i-1})

        Computing this value inside the beam loop repeats the same work
        many times, so training builds caches for the observed unigram,
        bigram, and trigram contexts. During decoding, ``_expand_states``
        first tries the trigram cache, then the bigram cache, then the
        unigram cache.

        The caches store log2 probabilities.

            ``_trans_logp_unigram[state]``
            ``_trans_logp_bigram[prev1][state]``
            ``_trans_logp_trigram[(prev2, prev1)][state]``

        Bigram cache entries include the unigram and bigram interpolation
        terms. Trigram cache entries include all three terms. The
        parenthesization in the probability calculation is kept stable so
        cached values match the previous per-call computation.
        """
        tag_unigrams = self._tag_unigrams
        tag_bigrams = self._tag_bigrams
        tag_trigrams = self._tag_trigrams

        lambda1, lambda2, lambda3 = self._lambda1, self._lambda2, self._lambda3
        inv_total_N = _safe_inverse(self._num_tag_tokens)

        # Computed once and reused below so the bigram and trigram caches
        # don't recompute the unigram contribution per (history, current).
        unigram_part = {
            state: lambda1 * (count * inv_total_N)
            for state, count in tag_unigrams.items()
        }
        self._trans_logp_unigram = {
            state: _safe_log2(p) for state, p in unigram_part.items()
        }

        bigram_logp = {}
        for prev1 in tag_bigrams.conditions():
            bigram_dist = tag_bigrams[prev1]
            inv_bigram_N = _safe_inverse(bigram_dist.N())

            bigram_logp[prev1] = {
                current: _safe_log2(
                    unigram_part.get(current, 0.0) + lambda2 * count * inv_bigram_N
                )
                for current, count in bigram_dist.items()
            }
        self._trans_logp_bigram = bigram_logp

        trigram_logp = {}
        for prev_pair in tag_trigrams.conditions():
            _, prev1 = prev_pair

            trigram_dist = tag_trigrams[prev_pair]
            bigram_dist = tag_bigrams[prev1]

            inv_trigram_N = _safe_inverse(trigram_dist.N())
            inv_bigram_N = _safe_inverse(bigram_dist.N())
            bigram_get = bigram_dist.get

            trigram_logp[prev_pair] = {
                current: _safe_log2(
                    unigram_part.get(current, 0.0)
                    + lambda2 * bigram_get(current, 0) * inv_bigram_N
                    + lambda3 * count * inv_trigram_N
                )
                for current, count in trigram_dist.items()
            }
        self._trans_logp_trigram = trigram_logp

    def _build_suffix_model(self):
        """
        Build the suffix model used for unseen words.

        The model stores three decode-time values: capitalization-split
        suffix tries, the successive-abstraction weight theta, and raw
        tag priors collapsed across capitalization.

        EOS is excluded from the priors because it is a sequence marker,
        not a lexical tag. Suffix counts come only from lexicon words
        with total count at most 10, following Brants's infrequent-word
        threshold for unknown-word modeling.
        """
        tag_unigrams = self._tag_unigrams
        word_tag_freqs = self._word_tag_freqs

        tag_counts: dict = {}
        for (tag, _), count in sorted(tag_unigrams.items()):
            # Exclude the EOS sentinel from all capitalization states.
            if tag == _EOS[0]:
                continue
            tag_counts[tag] = tag_counts.get(tag, 0) + count

        total = sum(tag_counts.values())
        if total > 0:
            tag_prior_probs = {tag: count / total for tag, count in tag_counts.items()}
        else:
            tag_prior_probs = {}

        # Theta controls how strongly the suffix recursion is pulled back
        # toward the less specific estimate at each abstraction step.
        priors = list(tag_prior_probs.values())
        n_priors = len(priors)
        if n_priors > 1:
            mean = sum(priors) / n_priors
            theta = (
                sum((prior - mean) ** 2 for prior in priors) / (n_priors - 1)
            ) ** 0.5
        else:
            theta = 0.0

        # Each capitalization bucket stores every suffix length up to 10,
        # so decode can walk from the shortest matched ending to the longest.
        suffix_trie_by_cap = {
            False: ConditionalFreqDist(),
            True: ConditionalFreqDist(),
        }

        # Sorted iteration makes suffix-trie insertion order-independent.
        for word in sorted(word_tag_freqs.conditions()):
            tag_freqs = word_tag_freqs[word]
            if not word or tag_freqs.N() > 10:
                continue

            suffix_trie = suffix_trie_by_cap[word[0].isupper()]
            max_suffix_len = min(len(word), 10)

            for m in range(1, max_suffix_len + 1):
                suffix_dist = suffix_trie[word[-m:]]
                for tag, count in sorted(tag_freqs.items()):
                    suffix_dist[tag] += count

        self._tag_prior_probs = tag_prior_probs
        self._theta = theta
        self._suffix_trie_by_cap = suffix_trie_by_cap

    def _unknown_tag_scores(self, word):
        """
        Score candidate tags for an unknown word using Brants's suffix
        model.

        The intuition is that a word's last few characters predict its
        tag well, since `-able` words tend to be adjectives, `-ing`
        words tend to be participles, and so on. Starting from the
        unigram tag prior, we walk one suffix character at a time up
        to the longest suffix we saw during training (capped at 10
        characters), blending each suffix's tag distribution into the
        running estimate via the successive abstraction recursion

            P(t | l_{n-i+1}...l_n) = (P_hat + theta * P_prev) / (1 + theta)

        If the word's tail is unfamiliar, the recursion never gets past
        the prior, which is the back-off case.

        :return: Bayes-inverted scores. Each raw tag maps to a quantity
                proportional to P(suffix | t). The P(suffix) constant
                drops out because it does not depend on the tag, which
                preserves the argmax without computing it. Tags with
                zero prior are omitted.
        """
        tag_priors = self._tag_prior_probs
        if not tag_priors:
            return {}

        is_capitalized = bool(word) and word[0].isupper()
        suffix_trie = self._suffix_trie_by_cap[is_capitalized]
        max_suffix_len = min(len(word), 10)

        # The trie contains every suffix length up to the cutoff. Once the
        # longest matching suffix is found, all shorter suffixes are also
        # available for successive abstraction.
        longest = 0
        for m in range(max_suffix_len, 0, -1):
            if word[-m:] in suffix_trie:
                longest = m
                break

        # No matched suffix means the estimate stays at the unigram prior.
        # After Bayes inversion, every tag then receives the same score.
        if longest == 0:
            return {tag: 1.0 for tag, prior in tag_priors.items() if prior > 0}

        theta = self._theta

        # With theta equal to zero there is no smoothing, so the estimate
        # is the empirical distribution of the longest matched suffix.
        # Tags absent from that bucket would have score zero, so they are
        # dropped from the result instead of being emitted as floor-only
        # candidates that the beam would prune anyway.
        if theta == 0.0:
            suffix_dist = suffix_trie[word[-longest:]]
            inv_suffix_N = 1.0 / suffix_dist.N()
            return {
                tag: (count * inv_suffix_N) / tag_priors[tag]
                for tag, count in suffix_dist.items()
                if tag_priors.get(tag, 0) > 0
            }

        # Dense successive abstraction updates every tag at every suffix
        # length. For tags absent from the current suffix bucket, that update
        # is the same shared shrinkage. Factor that shared term into one
        # scalar, and keep only tag specific corrections in delta.
        denom = 1.0 + theta
        miss_scale = theta / denom

        global_scale = 1.0
        delta: dict = {}

        for i in range(1, longest + 1):
            suffix_dist = suffix_trie[word[-i:]]
            inv_suffix_N = 1.0 / suffix_dist.N()

            # Apply the shared shrinkage for all tags, then add the suffix
            # evidence only for tags observed in this bucket.
            global_scale *= miss_scale
            corr_scale = inv_suffix_N / (denom * global_scale)

            for tag, count in suffix_dist.items():
                delta[tag] = delta.get(tag, 0.0) + count * corr_scale

        # In the factored form, Bayes inversion becomes
        #   P(t | suffix) / P(t) = global_scale * (1 + delta[t] / P(t)).
        # Untouched tags share one score. Touched tags get a correction
        # relative to their unigram prior.
        result = {}
        for tag, prior in tag_priors.items():
            if prior <= 0:
                continue
            extra = delta.get(tag)
            if extra is None:
                result[tag] = global_scale
            else:
                result[tag] = global_scale * (1.0 + extra / prior)

        return result

    def _tag_segmented(self, tokens):
        """
        Tag ``tokens`` as one or more sentences split on ``[.!?;]``.

        Each sentence-final punctuation token stays with the segment it
        closes, and a trailing fragment without sentence-final punctuation
        is still tagged as its own segment.
        """
        tagged: list = []
        segment: list = []

        sent_marks = _SENT_MARKS
        tagword = self._tagword
        pair_decoded = self._pair_decoded
        extend = tagged.extend

        for token in tokens:
            segment.append(token)
            if token in sent_marks:
                extend(pair_decoded(segment, tagword(segment)))
                segment.clear()

        if segment:
            extend(pair_decoded(segment, tagword(segment)))

        return tagged

    def _tagword(self, sent):
        """
        Tag one sentence with second-order Viterbi decoding.

        The lattice state is the last two tag states, so paths that share
        ``(state_{i-1}, state_i)`` are merged immediately. Known words draw
        candidates from the lexicon. Unknown words are scored either by the
        external ``unk`` tagger or by the suffix model. After each word, the
        beam keeps only states whose score is within ``log2(N)`` of the best
        surviving path. The decode then scores an explicit EOS transition
        using deleted interpolation over the unigram, bigram, and trigram
        EOS counts, and walks backpointers to recover the best state
        sequence.

        :param sent: words to tag
        :type sent: list[str]
        :return: list shaped ``[BOS, BOS, state_0, ..., state_{T-1}]``
                where each state is a ``(tag, capitalization)`` pair.
        """
        if not sent:
            return [_BOS, _BOS]

        T = len(sent)

        # Local bindings keep the hot loop on plain locals rather than
        # repeated attribute lookups.
        word_tag_freqs = self._word_tag_freqs
        tag_unigrams = self._tag_unigrams
        trans_logp_unigram_get = self._trans_logp_unigram.get
        log2_beam_threshold = self._log2_beam_threshold
        cap_on = self._use_capitalization
        unk = self._unk
        cache = self._candidate_tags_cache
        unknown_tag_scores = self._unknown_tag_scores
        expand_states = self._expand_states

        # Each level keeps only the best path reaching a given
        # ``(state_{i-1}, state_i)`` key. The backpointer stores
        # ``state_{i-2}`` so the best path can be reconstructed at the end.
        states = {(_BOS, _BOS): (0.0, _BOS)}
        state_history = [states]

        candidate_tags: tuple
        for word in sent:
            c_i = cap_on and bool(word) and word[0].isupper()
            tag_freqs = word_tag_freqs.get(word)

            if tag_freqs is not None:
                self.known += 1
            else:
                self.unknown += 1

            # External unknown-word taggers are treated as potentially
            # stateful. The built-in known-word and suffix-model paths are
            # pure given ``(word, c_i)`` and the trained model, so they cache.
            if tag_freqs is None and unk is not None:
                # Validate length before destructuring so a misbehaving
                # external tagger surfaces a clear error instead of a
                # cryptic "too many values to unpack" message.
                unk_out = list(unk.tag([word]))
                if len(unk_out) != 1:
                    raise ValueError(
                        f"unk tagger returned {len(unk_out)} tags for 1 word; "
                        f"expected exactly 1"
                    )
                ((_word, tag),) = unk_out
                state_i = (tag, c_i)
                unigram_logp = trans_logp_unigram_get(state_i, _LOG_FLOOR_2)
                candidate_tags = ((state_i, 0.0, unigram_logp),)
            else:
                cache_key = (word, c_i)
                cached = cache.get(cache_key)

                if cached is None:
                    if tag_freqs is not None:
                        # Known words only consider tags actually seen with
                        # that surface form. The lexical term is P(word | tag).
                        entries = []
                        for tag, tag_count in sorted(tag_freqs.items()):
                            state_i = (tag, c_i)
                            unigram_state_i = tag_unigrams[state_i]
                            unigram_logp = trans_logp_unigram_get(state_i, _LOG_FLOOR_2)
                            entries.append(
                                (
                                    state_i,
                                    log2(tag_count / unigram_state_i),
                                    unigram_logp,
                                )
                            )
                        candidate_tags = tuple(entries)
                    else:
                        # Unknown words use the suffix model as their lexical
                        # score. Bayes inversion turns the suffix posterior into
                        # the emission-like quantity used by the decoder.
                        suffix_scores = unknown_tag_scores(word)

                        if not suffix_scores:
                            # An untrained tagger has no suffix priors, so the
                            # only safe fallback is a literal ``Unk`` state.
                            state_i = ("Unk", c_i)
                            unigram_logp = trans_logp_unigram_get(state_i, _LOG_FLOOR_2)
                            candidate_tags = ((state_i, 0.0, unigram_logp),)
                        else:
                            entries = []
                            for tag, score in suffix_scores.items():
                                state_i = (tag, c_i)
                                unigram_logp = trans_logp_unigram_get(
                                    state_i, _LOG_FLOOR_2
                                )
                                entries.append(
                                    (state_i, _safe_log2(score), unigram_logp)
                                )
                            candidate_tags = tuple(entries)

                    cache[cache_key] = candidate_tags
                else:
                    candidate_tags = cached

            new_states, best_logp = expand_states(states, candidate_tags)

            # Threshold pruning keeps the beam relative to the best current
            # path, which is the pruning rule described in the paper.
            cutoff = best_logp - log2_beam_threshold
            states = {k: v for k, v in new_states.items() if v[0] >= cutoff}
            state_history.append(states)

        # EOS uses raw probabilities (not the logged cache) so the lambdas
        # can interpolate across orders before ``log2``. Scored once per
        # sentence, so skipping the cache is irrelevant for performance.
        tag_bigrams = self._tag_bigrams
        tag_trigrams = self._tag_trigrams
        lambda1, lambda2, lambda3 = self._lambda1, self._lambda2, self._lambda3
        num_tag_tokens = self._num_tag_tokens
        p_eos_unigram = (tag_unigrams[_EOS] / num_tag_tokens) if num_tag_tokens else 0.0

        best_final_key = next(iter(states))
        best_final_logp = float("-inf")

        for predecessor_key, (prefix_logp, _) in states.items():
            state_i_minus_1 = predecessor_key[1]

            # ``ConditionalFreqDist.__getitem__`` would create empty entries
            # for unseen keys, mutating the trained model during decode. Use
            # ``.get`` so an unseen history just contributes zero.
            bigram_dist = tag_bigrams.get(state_i_minus_1)
            if bigram_dist is None:
                p_eos_bigram = 0.0
            else:
                bigram_N = bigram_dist.N()
                p_eos_bigram = (bigram_dist[_EOS] / bigram_N) if bigram_N else 0.0

            trigram_dist = tag_trigrams.get(predecessor_key)
            if trigram_dist is None:
                p_eos_trigram = 0.0
            else:
                trigram_N = trigram_dist.N()
                p_eos_trigram = (trigram_dist[_EOS] / trigram_N) if trigram_N else 0.0

            p_eos_given_history = (
                lambda1 * p_eos_unigram
                + lambda2 * p_eos_bigram
                + lambda3 * p_eos_trigram
            )
            final_logp = prefix_logp + _safe_log2(p_eos_given_history)
            if final_logp > best_final_logp:
                best_final_logp = final_logp
                best_final_key = predecessor_key

        # Walking the stored ``state_{i-2}`` backpointers recovers the best
        # full state sequence from the best final state pair.
        states_reversed = [best_final_key[1]]
        if T >= 2:
            states_reversed.append(best_final_key[0])

        current_key = best_final_key
        for level in range(T, 2, -1):
            backpointer = state_history[level][current_key][1]
            states_reversed.append(backpointer)
            current_key = (backpointer, current_key[0])

        states_reversed.reverse()
        return [_BOS, _BOS] + states_reversed

    def _expand_states(self, states, candidate_tags):
        """
        Take one Viterbi step. For every predecessor state, score each
        candidate `state_i` from `candidate_tags` and accumulate the
        results into a new state dict keyed by
        `(state_i_minus_1, state_i)`. When two predecessors land on the
        same key, keep the higher-scoring one and discard the other. The
        second-order Markov assumption means everything after this point
        depends only on the last two states, so the discarded path can
        never beat the kept one. Ties (equal `path_logp`) are broken by
        keeping the first-encountered path; sorted candidate construction
        upstream makes that tie-break deterministic.

        `candidate_tags` is a sequence of `(state_i, log_emit,
        unigram_logp)` triples. `log_emit` is the lexical log-probability
        for known words, the Bayes-inverted suffix score for unknown
        words, or zero for an external unknown-word tagger.

        `unigram_logp` is the precomputed unigram-tier transition
        log-probability for `state_i`, used as the fallback when the
        trigram and bigram caches both miss. Precomputing it once per
        candidate removes a hot lookup from the inner loop on high-OOV
        corpora.

        :return: ``(new_states, best_logp)``. ``new_states`` maps
                 `(state_i_minus_1, state_i)` to
                 `(logp, state_i_minus_2)`, where `state_i_minus_2` is
                 the backpointer used to reconstruct the best path after
                 the final EOS step. `best_logp` is the maximum `logp`
                 across `new_states`, returned so the caller can apply
                 threshold pruning without a second pass.
        """
        trans_logp_trigram = self._trans_logp_trigram
        trans_logp_bigram = self._trans_logp_bigram

        new_states: dict = {}
        new_states_get = new_states.get
        best_logp = float("-inf")

        for predecessor_key, (prefix_logp, _) in states.items():
            state_i_minus_2, state_i_minus_1 = predecessor_key

            # Per-history caches are constant across the candidate loop;
            # resolving them once and using ``_EMPTY_DICT`` for misses
            # lets the inner ``.get`` calls run without ``None`` branches.
            trigram_logp = trans_logp_trigram.get(predecessor_key, _EMPTY_DICT)
            bigram_logp = trans_logp_bigram.get(state_i_minus_1, _EMPTY_DICT)
            trigram_logp_get = trigram_logp.get
            bigram_logp_get = bigram_logp.get

            for state_i, log_emit, unigram_logp in candidate_tags:
                trans_logp = trigram_logp_get(state_i)
                if trans_logp is None:
                    trans_logp = bigram_logp_get(state_i)
                    if trans_logp is None:
                        # The candidate tuple carries the unigram fallback,
                        # including the floor for states never observed as
                        # unigrams, so the inner loop avoids a third dict lookup.
                        trans_logp = unigram_logp

                # Parens match the previous ``step_logp + prefix_logp``
                # order so the sum is bit-identical to the old decoder.
                path_logp = prefix_logp + (trans_logp + log_emit)
                next_state = (state_i_minus_1, state_i)

                # When two paths land on the same (state_{i-1}, state_i)
                # key, only the higher-scoring prefix can matter going
                # forward: future transitions depend on the key alone.
                #
                # Underspecified by Brants: on exact ties, strict ``>``
                # keeps the first-encountered hypothesis. Sorted
                # candidate construction upstream fixes the iteration
                # order, so the tie-break is deterministic across reruns.
                # In practice this branch never fires (0 hits in 100k+
                # Treebank comparisons), so ``>`` vs ``>=`` is a no-op.
                prev_best = new_states_get(next_state)
                if prev_best is None or path_logp > prev_best[0]:
                    new_states[next_state] = (path_logp, state_i_minus_2)
                    if path_logp > best_logp:
                        best_logp = path_logp

        return new_states, best_logp

    @staticmethod
    def _pair_decoded(words, states):
        """Convert `_tagword` output into ``(word, tag)`` pairs by dropping
        the two BOS entries and the capitalization flag from each state."""
        return [(word, states[i + 2][0]) for i, word in enumerate(words)]


# -----------------------------
# Sentence segmentation helpers
# -----------------------------


def basic_sent_chop(data, raw=True, sent_marks=_SENT_MARKS):
    """
    Split a flat token sequence into sentence-like segments.

    Splits after tokens whose surface form appears in ``sent_marks``.
    Works for raw tokens and tagged ``(word, tag)`` tokens.

    :param data: flat sequence of tokens or ``(word, tag)`` tuples
    :param raw: whether ``data`` contains raw tokens
    :param sent_marks: token strings that end a sentence segment
    :return: list of sentence segments with the same token shape as input
    """
    sent_marks = set(sent_marks)
    sentences = []
    sentence = []

    for token in data:
        sentence.append(token)
        word = token if raw else token[0]
        if word in sent_marks:
            sentences.append(sentence)
            sentence = []

    if sentence:
        sentences.append(sentence)

    return sentences


# ------------
# Demo helpers
# ------------


def _treebank_demo_split(test_size=1000):
    """Return train, held-out final Treebank slice, and train vocabulary."""
    from nltk.corpus import treebank

    sents = list(treebank.tagged_sents())
    if not 0 < test_size < len(sents):
        raise ValueError(
            f"test_size must be between 1 and {len(sents) - 1}, got {test_size!r}"
        )

    cut = len(sents) - test_size
    train_sents = sents[:cut]
    test_sents = sents[cut:]
    train_vocab = {word for sent in train_sents for word, _ in sent}
    return train_sents, test_sents, train_vocab


def _score_tagger(tagger, test_sents, train_vocab):
    """Return overall, seen, and OOV scores for tagged sentences."""
    correct = total = 0
    seen_correct = seen_total = 0
    oov_correct = oov_total = 0

    for sent in test_sents:
        words = [word for word, _ in sent]
        gold = [tag for _, tag in sent]
        pred = [tag for _, tag in tagger.tag(words)]

        if len(pred) != len(gold):
            raise ValueError(f"tagger returned {len(pred)} tags for {len(gold)} tokens")

        for word, guess, truth in zip(words, pred, gold):
            total += 1
            hit = guess == truth
            if hit:
                correct += 1

            if word in train_vocab:
                seen_total += 1
                if hit:
                    seen_correct += 1
            else:
                oov_total += 1
                if hit:
                    oov_correct += 1

    return {
        "accuracy": correct / total if total else 0.0,
        "seen_accuracy": seen_correct / seen_total if seen_total else None,
        "oov_accuracy": oov_correct / oov_total if oov_total else None,
        "oov_rate": oov_total / total if total else 0.0,
        "total": total,
        "seen_total": seen_total,
        "oov_total": oov_total,
    }


def _format_score(score):
    return "n/a" if score is None else f"{score:.4f}"


# -----
# Demos
# -----


def demo(test_size=1000):
    """Evaluate TnT on a held-out Treebank slice."""
    train_sents, test_sents, train_vocab = _treebank_demo_split(test_size)

    for use_capitalization in (False, True):
        tagger = TnT(N=1000, C=use_capitalization)
        tagger.train(train_sents)
        scores = _score_tagger(tagger, test_sents, train_vocab)

        print(f"Capitalization: {use_capitalization}")
        print(f"Accuracy:       {scores['accuracy']:.4f}")
        print(f"Seen accuracy:  {_format_score(scores['seen_accuracy'])}")
        print(f"OOV accuracy:   {_format_score(scores['oov_accuracy'])}")
        print(f"OOV rate:       {scores['oov_rate']:.4f}")
        print()


def demo_errors(limit=25, test_size=1000):
    """Print the first ``limit`` errors from a held-out Treebank slice."""
    if limit < 1:
        raise ValueError(f"limit must be at least 1, got {limit!r}")

    train_sents, test_sents, train_vocab = _treebank_demo_split(test_size)

    tagger = TnT(N=1000, C=True)
    tagger.train(train_sents)

    shown = 0
    for sent in test_sents:
        words = [word for word, _ in sent]
        gold = [tag for _, tag in sent]
        pred = [tag for _, tag in tagger.tag(words)]

        if len(pred) != len(gold):
            raise ValueError(f"tagger returned {len(pred)} tags for {len(gold)} tokens")

        for word, guess, truth in zip(words, pred, gold):
            if guess == truth:
                continue

            status = "seen" if word in train_vocab else "OOV"
            print(f"{word!r} ({status}): guessed {guess!r}, gold {truth!r}")
            shown += 1
            if shown >= limit:
                return

    if shown == 0:
        print("No errors found.")



"""Regression tests for the quadratic-backtracking ReDoS in NLTK's Alpino corpus
reader (``nltk.corpus.reader.bracket_parse.AlpinoCorpusReader._normalize``) --
CWE-1333.

The node-normalization helper turned each Alpino ``<node ...>`` element into
s-expression notation with substitutions that chained several lazy ``.*?`` groups
hunting for trailing literals (``pos="``, ``word="``, ``/>``). On a single long
``<node ...`` line that contained the early anchors but never the trailing
literals, every lazy group rescanned the rest of the line and the engine
redistributed the match across the groups -- O(n^2) in the line length. The
public ``words()``/``tagged_words()``/``tagged_sents()``/``parsed_sents()`` entry
points reach this helper, so a crafted Alpino file could pin a CPU core. Each
``<node>`` is now parsed by extracting its attributes in a single linear scan;
ordinary corpora are read identically.
"""

import multiprocessing
import os
import sys
import traceback

from nltk.corpus.reader.bracket_parse import AlpinoCorpusReader

_SAMPLE = (
    '<alpino_ds version="1.3">\n'
    '  <node begin="0" cat="top" end="3" id="0" rel="top">\n'
    '    <node begin="0" cat="smain" end="3" id="1" rel="--">\n'
    '      <node begin="1" end="2" pos="noun" rel="su" word="hond"/>\n'
    '      <node begin="0" end="1" pos="det" rel="det" word="de"/>\n'
    '      <node begin="2" end="3" pos="verb" rel="hd" word="blaft"/>\n'
    "    </node>\n"
    "  </node>\n"
    "  <sentence>de hond blaft</sentence>\n"
    "</alpino_ds>\n"
)


def _reader(tmp_path, text):
    (tmp_path / "alpino.xml").write_text(text, encoding="ISO-8859-1")
    return AlpinoCorpusReader(str(tmp_path))


def test_alpino_reading_preserved(tmp_path):
    """An ordinary Alpino sentence is read exactly as before."""
    r = _reader(tmp_path, _SAMPLE)
    # words()/tagged_words() use the ordered=True path (sorted by 'begin').
    assert list(r.words()) == ["de", "hond", "blaft"]
    assert list(r.tagged_words()) == [
        ("de", "det"),
        ("hond", "noun"),
        ("blaft", "verb"),
    ]
    assert list(r.tagged_sents()) == [
        [("de", "det"), ("hond", "noun"), ("blaft", "verb")]
    ]
    # parsed_sents() uses the ordered=False path (XML order, nested tree).
    assert [str(t) for t in r.parsed_sents()] == [
        "(top (smain (noun hond) (det de) (verb blaft)))"
    ]


def test_alpino_node_without_pos_or_word_is_skipped(tmp_path):
    """A leaf node missing pos/word contributes nothing, as with the old regex."""
    text = (
        '<alpino_ds version="1.3">\n'
        '  <node begin="0" cat="top" end="1" id="0" rel="top">\n'
        '    <node begin="0" end="1" rel="su"/>\n'
        '    <node begin="0" end="1" pos="noun" rel="hd" word="kat"/>\n'
        "  </node>\n"
        "</alpino_ds>\n"
    )
    r = _reader(tmp_path, text)
    assert list(r.words()) == ["kat"]
    assert list(r.tagged_words()) == [("kat", "noun")]


def test_alpino_malformed_fields_left_unconverted(tmp_path):
    """Nodes whose fields break the old regex shape are left unconverted.

    The old substitutions required ``begin="(\\d+)"`` and ``pos="(\\w+)"``; a
    node with a non-numeric ``begin`` or a ``pos`` containing non-word chars was
    never converted (so it does not appear in the output). The attribute-based
    normalizer keeps those constraints for byte-for-byte compatibility.
    """
    text = (
        '<alpino_ds version="1.3">\n'
        '  <node begin="0" cat="top" end="3" id="0" rel="top">\n'
        '    <node begin="1a" end="2" pos="noun" rel="su" word="hond"/>\n'
        '    <node begin="0" end="1" pos="de-t" rel="det" word="de"/>\n'
        '    <node begin="2" end="3" pos="verb" rel="hd" word="blaft"/>\n'
        "  </node>\n"
        "</alpino_ds>\n"
    )
    r = _reader(tmp_path, text)
    # only the well-formed node survives (non-numeric begin / non-\w pos dropped)
    assert list(r.words()) == ["blaft"]
    assert list(r.tagged_words()) == [("blaft", "verb")]


def _words_worker(root):
    """Read words() from a crafted alpino corpus; exit 0 on success, 3 on error.

    On error the traceback is printed (and stderr flushed, since ``os._exit``
    skips normal shutdown) before exiting so a failure here is diagnosable in CI
    -- the parent process only sees the numeric exit code.
    """
    try:
        list(AlpinoCorpusReader(root).words())
        os._exit(0)
    except BaseException:
        traceback.print_exc()
        sys.stderr.flush()
        os._exit(3)


def test_alpino_normalize_is_linear_not_quadratic(tmp_path):
    """A single long, malformed ``<node ...`` line must not blow up quadratically.

    Run in a spawned process with a hard deadline: the linear normalizer returns
    in milliseconds, while the previous chained-lazy regex is O(n^2) and needs
    well over a minute at this size, so a regression is terminated and fails the
    test instead of pinning a core for the rest of the suite.
    """
    # Early anchors (begin=, many pos=) but never the required word="/ />.
    body = (
        '<alpino_ds version="1.3">\n  <node begin="1" '
        + 'pos="a" ' * 200_000
        + "\n</alpino_ds>\n"
    )
    (tmp_path / "alpino.xml").write_text(body, encoding="ISO-8859-1")

    ctx = multiprocessing.get_context("spawn")
    proc = ctx.Process(target=_words_worker, args=(str(tmp_path),))
    proc.start()
    proc.join(30)
    if proc.is_alive():
        proc.terminate()
        proc.join()
        raise AssertionError(
            "AlpinoCorpusReader.words() did not finish in time: ReDoS regressed"
        )
    assert proc.exitcode == 0, f"worker failed (exit {proc.exitcode})"



"""Push NLTK stopwords to nltk-data-hub/stopwords on HuggingFace.

One config per language, one parquet file per language.
The HF dataset viewer shows each language as a separate tab.

Usage:
    python push_stopwords.py <hf_token>
"""

import os
import shutil
import sys

import pandas as pd
from huggingface_hub import HfApi

REPO_ID = "nltk-data-hub/stopwords"

# BCP-47 codes for NLTK language names
LANG_CODES = {
    "albanian": "sq",
    "arabic": "ar",
    "azerbaijani": "az",
    "basque": "eu",
    "belarusian": "be",
    "bengali": "bn",
    "catalan": "ca",
    "chinese": "zh",
    "danish": "da",
    "dutch": "nl",
    "english": "en",
    "finnish": "fi",
    "french": "fr",
    "german": "de",
    "greek": "el",
    "hebrew": "he",
    "hinglish": "hi",
    "hungarian": "hu",
    "indonesian": "id",
    "italian": "it",
    "kazakh": "kk",
    "nepali": "ne",
    "norwegian": "no",
    "portuguese": "pt",
    "romanian": "ro",
    "russian": "ru",
    "slovene": "sl",
    "spanish": "es",
    "swedish": "sv",
    "tajik": "tg",
    "tamil": "ta",
    "turkish": "tr",
    "uzbek": "uz",
}

README_TEMPLATE = """\
---
language:
{lang_yaml}
configs:
{configs_yaml}
license: other
task_categories:
- text-classification
- token-classification
pretty_name: NLTK Stopwords
---

# NLTK Stopwords

Stopword lists from [NLTK](https://www.nltk.org/), covering {n_langs} languages.

Each language is a separate config. Each row is one stopword.

## Usage

```python
from datasets import load_dataset

# Load one language
ds = load_dataset("nltk-data-hub/stopwords", "portuguese")
words = ds["stopwords"]["word"]

# Load all languages
for lang in {lang_list_repr}:
    ds = load_dataset("nltk-data-hub/stopwords", lang)
    print(lang, ds["stopwords"].num_rows)
```

## Schema

| Column | Type | Description |
|---|---|---|
| `word` | `string` | The stopword |

## Languages and word counts

| Language | BCP-47 | Count |
|---|---|---|
{lang_stats}

## Source

Originally distributed as part of `nltk.download('stopwords')`.
Converted to Parquet for use with the HuggingFace `datasets` library.

## Citation

```bibtex
@book{nltk,
  author    = {Bird, Steven and Klein, Ewan and Loper, Edward},
  title     = {Natural Language Processing with Python},
  publisher = {O'Reilly Media},
  year      = {2009},
  url       = {https://www.nltk.org/}
}
```
"""


def build_per_language(outdir):
    """Write one parquet file per language under outdir/<lang>/stopwords.parquet."""
    from nltk.corpus import stopwords as sw

    langs = sorted(sw.fileids())
    counts = {}
    for lang in langs:
        words = sw.words(lang)
        df = pd.DataFrame({"word": words})
        lang_dir = os.path.join(outdir, lang)
        os.makedirs(lang_dir, exist_ok=True)
        df.to_parquet(os.path.join(lang_dir, "stopwords.parquet"), index=False)
        counts[lang] = len(words)
        print(f"  {lang}: {len(words)} words")

    return langs, counts


def build_readme(langs, counts):
    lang_yaml = "\n".join(f"- {LANG_CODES.get(l, l)}" for l in langs)

    configs_yaml = "\n".join(
        f"- config_name: {lang}\n"
        f"  data_files:\n"
        f"  - split: stopwords\n"
        f"    path: data/{lang}/stopwords.parquet"
        for lang in langs
    )

    lang_stats = "\n".join(
        f"| {lang} | {LANG_CODES.get(lang, '?')} | {counts[lang]:,} |" for lang in langs
    )

    lang_list_repr = repr(langs)

    return (
        README_TEMPLATE.replace("{lang_yaml}", lang_yaml)
        .replace("{configs_yaml}", configs_yaml)
        .replace("{n_langs}", str(len(langs)))
        .replace("{lang_list_repr}", lang_list_repr)
        .replace("{lang_stats}", lang_stats)
    )


def main():
    if len(sys.argv) < 2:
        print("Usage: python push_stopwords.py <hf_token>")
        sys.exit(1)

    token = sys.argv[1]
    api = HfApi(token=token)

    outdir = "/tmp/nltk_stopwords_v2"
    if os.path.exists(outdir):
        shutil.rmtree(outdir)
    os.makedirs(outdir)

    print("Building per-language parquet files...")
    langs, counts = build_per_language(os.path.join(outdir, "data"))
    print(f"  {len(langs)} languages, {sum(counts.values()):,} total words")

    print("Building README.md...")
    readme = build_readme(langs, counts)
    readme_path = os.path.join(outdir, "README.md")
    open(readme_path, "w").write(readme)

    # Create / ensure repo exists
    print(f"\nCreating repo {REPO_ID}...")
    api.create_repo(repo_id=REPO_ID, repo_type="dataset", exist_ok=True)

    # Delete old flat parquet if present
    print("Removing old flat data/stopwords.parquet...")
    try:
        api.delete_file(
            path_in_repo="data/stopwords.parquet",
            repo_id=REPO_ID,
            repo_type="dataset",
        )
    except Exception:
        pass  # didn't exist, fine

    # Upload all per-language parquets
    print("Uploading per-language parquet files...")
    for lang in langs:
        local_path = os.path.join(outdir, "data", lang, "stopwords.parquet")
        api.upload_file(
            path_or_fileobj=local_path,
            path_in_repo=f"data/{lang}/stopwords.parquet",
            repo_id=REPO_ID,
            repo_type="dataset",
        )
        print(f"  uploaded {lang}")

    print("Uploading README.md...")
    api.upload_file(
        path_or_fileobj=readme_path,
        path_in_repo="README.md",
        repo_id=REPO_ID,
        repo_type="dataset",
    )

    print(f"\nDone: https://huggingface.co/datasets/{REPO_ID}")


if __name__ == "__main__":
    main()



import unittest
import warnings

from nltk.langnames import lang2q, langcode, langname, q2name, q2tag, tag2q


class TestTag2Q(unittest.TestCase):
    def test_known_tag(self):
        self.assertEqual(tag2q("nds-u-sd-demv"), "Q4289225")

    def test_unknown_tag(self):
        self.assertIsNone(tag2q("zzzz-unknown-tag"))

    def test_empty_string(self):
        self.assertIsNone(tag2q(""))

    def test_unknown_tag_strict(self):
        with self.assertRaises(LookupError):
            tag2q("zzzz-unknown-tag", strict=True)

    def test_known_tag_strict(self):
        self.assertEqual(tag2q("nds-u-sd-demv", strict=True), "Q4289225")


class TestQ2Tag(unittest.TestCase):
    def test_known_qcode(self):
        self.assertEqual(q2tag("Q4289225"), "nds-u-sd-demv")

    def test_unknown_qcode(self):
        self.assertIsNone(q2tag("Q0000000"))

    def test_empty_string(self):
        self.assertIsNone(q2tag(""))

    def test_unknown_qcode_strict(self):
        with self.assertRaises(LookupError):
            q2tag("Q0000000", strict=True)

    def test_known_qcode_strict(self):
        self.assertEqual(q2tag("Q4289225", strict=True), "nds-u-sd-demv")


class TestQ2Name(unittest.TestCase):
    def test_known_full(self):
        self.assertEqual(q2name("Q4289225"), "Low German: Mecklenburg-Vorpommern")

    def test_known_short(self):
        self.assertEqual(q2name("Q4289225", "short"), "Low German")

    def test_unknown_qcode(self):
        self.assertIsNone(q2name("Q0000000"))

    def test_empty_string(self):
        self.assertIsNone(q2name(""))

    def test_unknown_qcode_strict(self):
        with self.assertRaises(LookupError):
            q2name("Q0000000", strict=True)

    def test_known_qcode_strict(self):
        self.assertEqual(
            q2name("Q4289225", strict=True), "Low German: Mecklenburg-Vorpommern"
        )


class TestLang2Q(unittest.TestCase):
    def test_known_language(self):
        self.assertEqual(lang2q("Low German"), "Q25433")

    def test_unknown_language(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.assertIsNone(lang2q("NonexistentLanguage"))

    def test_empty_string(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.assertIsNone(lang2q(""))

    def test_unknown_language_strict(self):
        with self.assertRaises(LookupError):
            lang2q("NonexistentLanguage", strict=True)

    def test_known_language_strict(self):
        self.assertEqual(lang2q("Low German", strict=True), "Q25433")


class TestLangName(unittest.TestCase):
    def test_known_tag(self):
        self.assertEqual(
            langname("ca-Latn-ES-valencia"), "Catalan: Latin: Spain: Valencian"
        )

    def test_known_tag_short(self):
        self.assertEqual(langname("ca-Latn-ES-valencia", typ="short"), "Catalan")

    def test_retired_code(self):
        self.assertEqual(langname("fri"), "Western Frisian")

    def test_unknown_code(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.assertIsNone(langname("zzz"))

    def test_none_input(self):
        self.assertIsNone(langname(None))

    def test_unknown_code_strict(self):
        with self.assertRaises(LookupError):
            langname("zzz", strict=True)

    def test_none_input_strict(self):
        with self.assertRaises(LookupError):
            langname(None, strict=True)

    def test_known_tag_strict(self):
        self.assertEqual(
            langname("ca-Latn-ES-valencia", strict=True),
            "Catalan: Latin: Spain: Valencian",
        )


class TestLangCode(unittest.TestCase):
    def test_known_name(self):
        self.assertEqual(langcode("Modern Greek (1453-)"), "el")

    def test_known_name_3letter(self):
        self.assertEqual(langcode("Modern Greek (1453-)", typ=3), "ell")

    def test_retired_name(self):
        self.assertEqual(langcode("Western Frisian"), "fy")

    def test_unknown_name(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.assertIsNone(langcode("NotARealLanguage"))

    def test_unknown_name_strict(self):
        with self.assertRaises(LookupError):
            langcode("NotARealLanguage", strict=True)

    def test_known_name_strict(self):
        self.assertEqual(langcode("Modern Greek (1453-)", strict=True), "el")


if __name__ == "__main__":
    unittest.main()



"""Regression tests for the unbounded-default-``max_len`` DoS in
``nltk.util.everygrams`` (CWE-770; CVE-2026-12861).

With ``max_len`` left at its ``-1`` default it expands to ``len(sequence)``, so
enumerating every n-gram of every length from ``min_len`` to ``len(sequence)``
over each window yields O(n**2) tuples totalling O(n**3) elements -- a
few-thousand-token sequence then allocates gigabytes and OOM-kills the process.
The defaulted ``max_len`` is now capped by ``MAX_EVERYGRAMS_DEFAULT_LEN``; a
longer sequence with the default raises ``ValueError`` asking for an explicit
``max_len``. An explicitly supplied ``max_len`` is never capped.

The "must not allocate" test runs in a spawned process with a hard timeout, and
the worker reports its outcome through its exit code (no queue/thread, so it is
robust on free-threaded builds), so a regression cannot OOM/hang the suite.
"""

import multiprocessing
import os

import pytest

from nltk.util import MAX_EVERYGRAMS_DEFAULT_LEN, everygrams


def test_max_everygrams_default_len_is_a_finite_positive_int():
    assert isinstance(MAX_EVERYGRAMS_DEFAULT_LEN, int)
    assert MAX_EVERYGRAMS_DEFAULT_LEN > 0


def test_behaviour_preserved_for_small_sequences():
    sent = "a b c".split()
    # The values asserted in the everygrams doctest must be unchanged.
    assert list(everygrams(sent)) == [
        ("a",),
        ("a", "b"),
        ("a", "b", "c"),
        ("b",),
        ("b", "c"),
        ("c",),
    ]
    assert list(everygrams(sent, max_len=2)) == [
        ("a",),
        ("a", "b"),
        ("b",),
        ("b", "c"),
        ("c",),
    ]
    assert list(everygrams(sent, min_len=2)) == [
        ("a", "b"),
        ("a", "b", "c"),
        ("b", "c"),
    ]


def test_default_allowed_up_to_the_cap():
    # A sequence exactly at the cap still uses the full default (no raise).
    toks = [str(i) for i in range(MAX_EVERYGRAMS_DEFAULT_LEN)]
    out = list(everygrams(toks))
    n = MAX_EVERYGRAMS_DEFAULT_LEN
    assert len(out) == n * (n + 1) // 2


def test_oversized_default_raises():
    # One past the cap with the default max_len is refused -- before allocating.
    toks = [str(i) for i in range(MAX_EVERYGRAMS_DEFAULT_LEN + 1)]
    with pytest.raises(ValueError):
        list(everygrams(toks))


def test_explicit_max_len_is_never_capped():
    # An explicit (bounded) max_len over a long sequence stays linear and allowed.
    toks = [str(i) for i in range(5000)]
    out = list(everygrams(toks, max_len=3))
    # 1-, 2- and 3-grams over 5000 tokens: ~3 * 5000, far below any cube.
    assert len(out) == 5000 + 4999 + 4998


_TIMEOUT = 60
# Worker outcomes, reported via exit code (avoids a result queue / feeder thread,
# which is fragile on free-threaded builds).
_EXIT_REFUSED = 0  # ValueError raised before any allocation (the expected result)
_EXIT_ENUMERATED = 2  # the unbounded enumeration ran to completion (regression)
_EXIT_OTHER = 3  # any other failure


def _everygrams_worker():
    # With the default max_len this would enumerate O(n**2) tuples totalling
    # O(n**3) elements if the guard were removed. Consume the iterator WITHOUT
    # materializing it into a list: each tuple is transient, so peak memory
    # stays low and a regression is caught by the parent's timeout rather than
    # risking an OS OOM kill that could destabilize the whole CI job.
    toks = [str(i) for i in range(2000)]
    try:
        for _ in everygrams(toks):
            pass
        os._exit(_EXIT_ENUMERATED)
    except ValueError:
        os._exit(_EXIT_REFUSED)
    except BaseException:
        os._exit(_EXIT_OTHER)


def test_oversized_default_does_not_allocate():
    """everygrams(long_seq) with the default max_len must be refused, not run."""
    ctx = multiprocessing.get_context("spawn")
    proc = ctx.Process(target=_everygrams_worker)
    proc.start()
    proc.join(_TIMEOUT)
    if proc.is_alive():
        proc.terminate()
        proc.join()
        raise AssertionError(
            "everygrams() did not return quickly -> unbounded O(n**3) allocation (DoS)"
        )
    assert proc.exitcode == _EXIT_REFUSED, (
        "everygrams() with the default max_len over a long sequence was not "
        f"refused before allocating (worker exit code {proc.exitcode}); "
        "expected a fast ValueError"
    )



"""
Regression tests for ``nltk.tag.crf.CRFTagger``.
"""

import pytest

pytest.importorskip("pycrfsuite")

from nltk.tag.crf import CRFTagger

_TRAIN = [
    [("the", "DT"), ("cat", "NN"), ("sat", "VBD")],
    [("a", "DT"), ("dog", "NN"), ("ran", "VBD")],
    [("the", "DT"), ("dog", "NN"), ("sat", "VBD")],
    [("a", "DT"), ("cat", "NN"), ("ran", "VBD")],
]

_SAMPLE_SENT = ["the", "cat", "sat"]
_TAGS = {"DT", "NN", "VBD"}


@pytest.mark.parametrize(
    ("method", "bad", "match"),
    [
        ("tag", "the cat sat", "list of tokens"),
        ("tag", b"the cat sat", "list of tokens"),
        ("tag_sents", "the cat sat", "tokenized sentences"),
        ("tag_sents", b"the cat sat", "tokenized sentences"),
        ("tag_sents", ["the", "cat", "sat"], "tokenized sentences"),
        ("tag_sents", ("the", "cat", "sat"), "tokenized sentences"),
    ],
)
def test_rejects_bad_input_shapes(method, bad, match):
    ct = CRFTagger()
    with pytest.raises(TypeError, match=match):
        getattr(ct, method)(bad)


def test_training_options_are_copied():
    opts = {"c1": 0.5, "c2": 1.0}
    ct = CRFTagger(training_opt=opts)

    opts["c1"] = 99.0

    assert ct._training_options == {"c1": 0.5, "c2": 1.0}


@pytest.mark.parametrize(
    ("token", "expected"),
    [
        (
            "University",
            ["CAPITALIZATION", "SUF_y", "SUF_ty", "SUF_ity", "WORD_University"],
        ),
        ("A1", ["CAPITALIZATION", "HAS_NUM", "SUF_1", "WORD_A1"]),
        ("...", ["PUNCTUATION", "SUF_.", "SUF_..", "WORD_..."]),
        ("", []),
    ],
)
def test_default_features_are_cached_as_tuples(token, expected):
    ct = CRFTagger()

    first = ct._get_features([token], 0)
    second = ct._get_features([token], 0)

    assert first == expected
    assert second == expected
    assert first is not second
    assert ct._feature_cache[token] == tuple(expected)


def test_custom_feature_function_bypasses_default_cache():
    def feature_func(tokens, idx):
        prev = "<BOS>" if idx == 0 else tokens[idx - 1]
        return [f"TOKEN={tokens[idx]}", f"PREV={prev}"]

    ct = CRFTagger(feature_func=feature_func)

    assert ct._feature_func is feature_func
    assert ct._feature_func(["a", "b"], 1) == ["TOKEN=b", "PREV=a"]
    assert ct._feature_cache == {}


def test_clear_feature_cache_drops_cached_entries():
    ct = CRFTagger()

    ct._get_features(["University"], 0)
    ct._get_features(["dog"], 0)
    assert ct._feature_cache

    ct.clear_feature_cache()
    assert ct._feature_cache == {}

    ct._get_features(["University"], 0)
    assert "University" in ct._feature_cache


def test_tag_sents_kwargs_compatibility():
    ct = CRFTagger()

    with pytest.warns(DeprecationWarning, match="sents=.*deprecated"):
        with pytest.raises(RuntimeError, match="No model file set"):
            ct.tag_sents(sents=[["a", "b"]])

    with pytest.raises(TypeError, match="both 'sentences' and 'sents'"):
        ct.tag_sents([["a"]], sents=[["b"]])

    with pytest.raises(TypeError, match="unexpected keyword"):
        ct.tag_sents([["a"]], extra=True)


def test_train_tag_round_trip(tmp_path):
    model_file = tmp_path / "model.crf.tagger"

    trained = CRFTagger()
    trained.train(_TRAIN, str(model_file))
    assert model_file.exists()

    tagged = trained.tag(_SAMPLE_SENT)
    assert [word for word, _ in tagged] == _SAMPLE_SENT
    assert all(tag in _TAGS for _, tag in tagged)

    reloaded = CRFTagger()
    reloaded.set_model_file(str(model_file))
    assert reloaded.tag(_SAMPLE_SENT) == tagged



"""Regression tests for the untrusted-search-path fix in ReppTokenizer (CWE-426).

``find_repptokenizer`` used ``os.path.exists(repp_dirname)`` to accept a directory
as-is. For a *relative* ``repp_dirname`` that resolves against the current working
directory and is checked *before* the ``REPP_TOKENIZER`` environment variable, so
an attacker who can write a ``./<name>/src/repp`` executable into the CWD could
have it run (the command path contains a separator, so ``subprocess`` executes it
directly) -- overriding a trusted ``REPP_TOKENIZER``.

Only an explicit *absolute* directory is now taken as-is; a relative name is
resolved through ``REPP_TOKENIZER``.
"""

import os
import pathlib

import pytest

from nltk.tokenize.repp import ReppTokenizer


def _make_repp_dir(path):
    (path / "src").mkdir(parents=True, exist_ok=True)
    (path / "erg").mkdir(parents=True, exist_ok=True)
    (path / "src" / "repp").write_bytes(b"")
    (path / "erg" / "repp.set").write_bytes(b"")
    return path


def test_relative_dirname_not_resolved_against_cwd(tmp_path, monkeypatch):
    """A relative repp_dirname matching a CWD directory must NOT be used."""
    cwd = tmp_path / "cwd"
    cwd.mkdir()
    _make_repp_dir(cwd / "reppdir")  # attacker-planted in the CWD
    monkeypatch.chdir(cwd)
    monkeypatch.delenv("REPP_TOKENIZER", raising=False)

    with pytest.raises(LookupError):
        ReppTokenizer("reppdir")


def test_cwd_does_not_override_configured_repp_tokenizer(tmp_path, monkeypatch):
    """A CWD directory must not shadow a configured REPP_TOKENIZER."""
    cwd = tmp_path / "cwd"
    cwd.mkdir()
    _make_repp_dir(cwd / "reppdir")  # attacker-planted in the CWD
    trusted = _make_repp_dir(tmp_path / "trusted")  # the configured location
    monkeypatch.chdir(cwd)
    monkeypatch.setenv("REPP_TOKENIZER", str(trusted))

    tok = ReppTokenizer("reppdir")
    assert os.path.realpath(tok.repp_dir) == os.path.realpath(
        str(trusted)
    ), "CWD directory overrode the trusted REPP_TOKENIZER"


def test_absolute_path_still_accepted(tmp_path, monkeypatch):
    """An explicit absolute directory is still used as-is, incl. as a Path."""
    abs_dir = _make_repp_dir(tmp_path / "absrepp")
    monkeypatch.delenv("REPP_TOKENIZER", raising=False)

    assert os.path.realpath(ReppTokenizer(str(abs_dir)).repp_dir) == os.path.realpath(
        str(abs_dir)
    )
    # a pathlib.Path is accepted too (normalised via os.fspath) and resolves to
    # the same directory.
    assert os.path.realpath(
        ReppTokenizer(pathlib.Path(abs_dir)).repp_dir
    ) == os.path.realpath(str(abs_dir))



# Natural Language Toolkit: HuggingFace dataset integration
#
# Copyright (C) 2001-2026 NLTK Project
# URL: <https://www.nltk.org/>
# For license information, see LICENSE.TXT

"""
HuggingFace datasets integration for NLTK.

Provides a PathPointer subclass that reads directly from the HuggingFace
datasets cache, and a ``download()`` function that populates that cache.

Usage::

    import nltk
    nltk.download('stopwords', hf=True)            # download to HF cache
    nltk.corpus.stopwords.words('portuguese')      # HF fallback if not in ~/nltk_data
    nltk.corpus.stopwords.words('portuguese', hf=True)  # HF cache directly

Registry schema
---------------
Each entry in ``REGISTRY`` must declare:

``repo`` (str)
    HuggingFace dataset repo id, e.g. ``"nltk-data-hub/stopwords"``.

``split`` (str)
    HF split name to load, e.g. ``"stopwords"``, ``"train"``.

``structure`` (str)
    How the corpus is organised on HF:

    ``"multi_config"``
        One HF config per NLTK fileid.  No assumption about what that
        dimension represents (language, category, author, etc.).
        ``fileid`` → config name.

    ``"flat"``
        Single config, flat table.  A ``fileid_column`` value is used
        to select rows for a given fileid.

    ``"single"``
        Single config, no sub-selection.  The whole split is the corpus.

``content_type`` (str)
    How rows are serialised to the byte/text stream NLTK readers expect:

    ``"word_list"``
        Each row is one entry; ``text_column`` holds the string.
        Serialised as one entry per line.

    ``"raw_text"``
        Rows have a ``text_column`` with full document text.  When a
        fileid is given, rows are filtered by ``fileid_column``.
        Serialised as the raw text string.

    (Add new types here as more corpora are onboarded.)

``cache_probe`` (str)
    A single parquet path inside the repo used to detect whether the
    corpus has already been downloaded locally.  No network request is
    made; ``huggingface_hub.try_to_load_from_cache`` inspects the local
    filesystem only.

Optional keys (required by certain content types):

``text_column``    column that holds the main text / word value.
``fileid_column``  column that identifies which NLTK fileid a row belongs to.
``label_column``   column for classification labels (future use).
"""

import io

# ---------------------------------------------------------------------------
# Registry
# ---------------------------------------------------------------------------

REGISTRY = {
    "stopwords": {
        "repo": "nltk-data-hub/stopwords",
        "split": "stopwords",
        "structure": "multi_config",
        "content_type": "word_list",
        "text_column": "word",
        "cache_probe": "data/english/stopwords.parquet",
    },
}


# ---------------------------------------------------------------------------
# Cache detection (no network)
# ---------------------------------------------------------------------------


def _is_cached(corpus_id):
    """Return True if the corpus parquet exists in the local HF datasets cache."""
    info = REGISTRY.get(corpus_id)
    if info is None:
        return False
    try:
        from huggingface_hub import try_to_load_from_cache

        result = try_to_load_from_cache(
            repo_id=info["repo"],
            filename=info["cache_probe"],
            repo_type="dataset",
        )
        return result is not None and result != "no_connection"
    except Exception:
        return False


# ---------------------------------------------------------------------------
# Content serialisation
# ---------------------------------------------------------------------------


def _serialise(ds, info, fileid=None):
    """
    Convert an HF dataset ``ds`` to the byte/text content that an NLTK
    corpus reader would find in a plain file.

    :param ds: ``datasets.Dataset`` already filtered/selected for this corpus.
    :param info: REGISTRY entry for the corpus.
    :param fileid: the NLTK fileid being requested (used by some types).
    :returns: str — file content as NLTK expects it.
    """
    content_type = info.get("content_type", "raw_text")

    if content_type == "word_list":
        return "\n".join(ds[info["text_column"]])

    if content_type == "raw_text":
        col = info["text_column"]
        texts = ds[col]
        return "\n".join(texts) if len(texts) > 1 else (texts[0] if texts else "")

    raise NotImplementedError(
        f"content_type={content_type!r} is not implemented. "
        "Add a handler in nltk.huggingface.dataset._serialise()."
    )


def _load_hf_dataset(info, fileid=None):
    """
    Load the appropriate HF dataset slice for *fileid*, respecting structure.

    :param info: REGISTRY entry.
    :param fileid: NLTK fileid (may be None for single-structure corpora).
    :returns: ``datasets.Dataset``.
    """
    from datasets import load_dataset

    structure = info.get("structure", "single")

    if structure == "multi_config":
        if fileid is None:
            raise ValueError(
                "fileid is required for multi_config corpora. "
                "Pass the config name (e.g. a language or category)."
            )
        return load_dataset(info["repo"], fileid, split=info["split"])

    if structure == "flat":
        ds = load_dataset(info["repo"], split=info["split"])
        if fileid is not None:
            col = info["fileid_column"]
            ds = ds.filter(lambda row: row[col] == fileid)
        return ds

    # single
    return load_dataset(info["repo"], split=info["split"])


# ---------------------------------------------------------------------------
# HFDatasetPathPointer
# ---------------------------------------------------------------------------


class HFDatasetPathPointer:
    """
    A ``PathPointer``-compatible object backed by a HuggingFace dataset
    stored in the local HF datasets cache (~/.cache/huggingface/datasets/).

    Satisfies the NLTK PathPointer interface (``open`` / ``file_size`` /
    ``join``) so that existing corpus readers work unchanged after
    ``nltk.download(..., hf=True)``.
    """

    def __init__(self, corpus_id, fileid=None):
        self.corpus_id = corpus_id
        self.fileid = fileid

    # -- PathPointer interface -----------------------------------------------

    def open(self, encoding=None):
        """Return a stream of file content as NLTK corpus readers expect."""
        info = REGISTRY[self.corpus_id]
        ds = _load_hf_dataset(info, fileid=self.fileid)
        content = _serialise(ds, info, fileid=self.fileid)
        if encoding:
            return io.StringIO(content)
        return io.BytesIO(content.encode("utf-8"))

    def file_size(self):
        return 0

    def join(self, fileid):
        return HFDatasetPathPointer(self.corpus_id, fileid)

    # -- fileids (duck-typed by find_corpus_fileids) -------------------------

    def fileids(self):
        """Return sorted list of fileids available for this corpus."""
        info = REGISTRY.get(self.corpus_id)
        if info is None or not _is_cached(self.corpus_id):
            return []
        structure = info.get("structure", "single")
        try:
            if structure == "multi_config":
                from datasets import get_dataset_config_names

                return sorted(get_dataset_config_names(info["repo"]))
            if structure == "flat":
                from datasets import load_dataset

                ds = load_dataset(info["repo"], split=info["split"])
                return sorted(ds.unique(info["fileid_column"]))
            return [info["split"]]  # single
        except Exception:
            return []

    # -- repr / path ---------------------------------------------------------

    @property
    def path(self):
        repo = REGISTRY.get(self.corpus_id, {}).get("repo", self.corpus_id)
        return f"hf://{repo}"

    def __str__(self):
        return f"{self.path}/{self.fileid}" if self.fileid else self.path

    def __repr__(self):
        return f"HFDatasetPathPointer({self.corpus_id!r}, {self.fileid!r})"


# Register as virtual subclass of PathPointer — avoids circular import at load
def _register_path_pointer():
    try:
        from nltk.data import PathPointer

        PathPointer.register(HFDatasetPathPointer)
    except Exception:
        pass


_register_path_pointer()


# ---------------------------------------------------------------------------
# Public API
# ---------------------------------------------------------------------------


def download(corpus_id, token=None, quiet=False):
    """
    Download an NLTK corpus from HuggingFace into the HF datasets cache
    (``~/.cache/huggingface/datasets/``).

    :param corpus_id: NLTK corpus id, e.g. ``'stopwords'``.
    :param token: optional HuggingFace API token for private repos.
    :param quiet: suppress progress output.
    :raises ValueError: if *corpus_id* is not in the HF registry.
    """
    info = REGISTRY.get(corpus_id)
    if info is None:
        raise ValueError(
            f"Corpus {corpus_id!r} is not available on HuggingFace.\n"
            f"Available: {sorted(REGISTRY)}"
        )

    from datasets import load_dataset

    kwargs = {"token": token} if token else {}
    structure = info.get("structure", "single")

    if structure == "multi_config":
        from datasets import get_dataset_config_names

        configs = get_dataset_config_names(info["repo"])
        result = {
            cfg: load_dataset(info["repo"], cfg, split=info["split"], **kwargs)
            for cfg in configs
        }
        if not quiet:
            total = sum(len(d) for d in result.values())
            print(
                f"[nltk_hf] '{corpus_id}' downloaded from {info['repo']} "
                f"({len(configs)} configs, {total:,} rows)"
            )
        return result

    else:  # flat or single
        ds = load_dataset(info["repo"], split=info["split"], **kwargs)
        if not quiet:
            print(
                f"[nltk_hf] '{corpus_id}' downloaded from {info['repo']} "
                f"({len(ds):,} rows)"
            )
        return ds


def load_data(corpus_id, fileid=None):
    """
    Load data for *corpus_id* directly from the HF datasets cache and return
    it as a string in the format NLTK corpus readers expect.

    :param corpus_id: NLTK corpus id, e.g. ``'stopwords'``.
    :param fileid: sub-resource identifier (config name, category, fileid, …).
    :returns: str.
    :raises LookupError: if the corpus is not in the registry or not cached.
    """
    info = REGISTRY.get(corpus_id)
    if info is None:
        raise LookupError(
            f"Corpus {corpus_id!r} is not in the HuggingFace NLTK registry."
        )
    if not _is_cached(corpus_id):
        raise LookupError(
            f"Corpus {corpus_id!r} not found in HF datasets cache. "
            f"Run: nltk.download({corpus_id!r}, hf=True)"
        )
    ds = _load_hf_dataset(info, fileid=fileid)
    return _serialise(ds, info, fileid=fileid)



import os
from pathlib import Path

import pytest

import nltk.parse.malt as malt
from nltk.parse.malt import MaltParser


def _minimal_malt_parser(tmp_path):
    model_dir = tmp_path / "model"
    model_dir.mkdir()
    model_file = model_dir / "model.mco"
    model_file.write_text("dummy model\n", encoding="utf-8")

    working_dir = tmp_path / "work"
    working_dir.mkdir()

    parser = MaltParser.__new__(MaltParser)
    parser._trained = True
    parser.model = str(model_file)
    parser.working_dir = str(working_dir)
    parser.malt_jars = ["dummy-malt.jar"]
    parser.additional_java_args = []
    return parser, model_dir


def _write_minimal_parse(cmd):
    output_path = Path(cmd[cmd.index("-o") + 1])
    output_path.write_text(
        "1\thello\t_\tNN\tNN\t_\t0\tnull\t_\t_\n",
        encoding="utf-8",
    )


def test_malt_parse_uses_subprocess_cwd_without_changing_process_cwd(
    monkeypatch, tmp_path
):
    parser, model_dir = _minimal_malt_parser(tmp_path)
    service_dir = tmp_path / "service"
    service_dir.mkdir()
    monkeypatch.chdir(service_dir)

    def forbidden_chdir(path):
        raise AssertionError(f"os.chdir() should not be called with {path!r}")

    monkeypatch.setattr(malt.os, "chdir", forbidden_chdir)

    captured = {}

    def fake_execute(cmd, verbose=False, cwd=None):
        captured["cwd"] = cwd
        captured["process_cwd"] = os.getcwd()
        captured["input_file"] = Path(cmd[cmd.index("-i") + 1])
        captured["output_file"] = Path(cmd[cmd.index("-o") + 1])
        _write_minimal_parse(cmd)
        return 0

    parser._execute = fake_execute

    graphs = list(parser.parse_tagged_sents([[("hello", "NN")]]))

    assert len(graphs) == 1
    assert captured["cwd"] == str(model_dir)
    assert captured["process_cwd"] == str(service_dir)
    assert os.getcwd() == str(service_dir)
    assert not captured["input_file"].exists()
    assert not captured["output_file"].exists()


def test_malt_parse_cleans_temp_files_and_preserves_cwd_on_execute_exception(
    monkeypatch, tmp_path
):
    parser, model_dir = _minimal_malt_parser(tmp_path)
    service_dir = tmp_path / "service"
    service_dir.mkdir()
    monkeypatch.chdir(service_dir)

    captured = {}

    def fake_execute(cmd, verbose=False, cwd=None):
        captured["cwd"] = cwd
        captured["process_cwd"] = os.getcwd()
        captured["input_file"] = Path(cmd[cmd.index("-i") + 1])
        captured["output_file"] = Path(cmd[cmd.index("-o") + 1])
        assert captured["input_file"].exists()
        assert captured["output_file"].exists()
        raise RuntimeError("forced parser failure")

    parser._execute = fake_execute

    with pytest.raises(RuntimeError, match="forced parser failure"):
        list(parser.parse_tagged_sents([[("hello", "NN")]]))

    assert captured["cwd"] == str(model_dir)
    assert captured["process_cwd"] == str(service_dir)
    assert os.getcwd() == str(service_dir)
    assert not captured["input_file"].exists()
    assert not captured["output_file"].exists()



"""Regression tests for ReDoS in TextTilingTokenizer (CWE-1333).

``_mark_paragraph_breaks`` scans the input for blank-line paragraph breaks with
``[ \\t\\r\\f\\v]*\\n[ \\t\\r\\f\\v]*\\n[ \\t\\r\\f\\v]*``. With the plain greedy
``re`` pattern, ``finditer`` rescans a long horizontal-whitespace run that has no
blank line quadratically, so a crafted whitespace blob hangs ``tokenize()``. The
pattern now uses possessive quantifiers (regex module), making the scan linear.
The whitespace class does not overlap ``"\\n"``, so the matches are unchanged.

The "must not hang" tests run the work in a separate process (spawn) with a hard
timeout and ``terminate()`` on overrun, so a regression to a quadratic regex
cannot keep burning CPU for the rest of the suite, and any exception in the
worker is propagated back to the assertions instead of being swallowed.
"""

import multiprocessing
import queue

from nltk.tokenize.texttiling import TextTilingTokenizer

# A long horizontal-whitespace run with no blank line: ~256 KB. Linear with the
# possessive pattern (sub-millisecond); ~quadratic and tens of seconds with the
# old greedy one.
_CRAFTED_TEXT = " \t" * 128_000
_TIMEOUT = 15

# stopwords are passed explicitly so constructing the tokenizer needs no corpus
# download; the vulnerable scan is in _mark_paragraph_breaks, before any
# stopword use.
_STOPWORDS = ["the", "a", "of", "and", "to"]


def _tokenizer():
    return TextTilingTokenizer(stopwords=_STOPWORDS)


def _mark_worker(result_q):
    try:
        result_q.put(("ok", _tokenizer()._mark_paragraph_breaks(_CRAFTED_TEXT)))
    except BaseException as exc:  # surface to the parent process
        result_q.put(("error", repr(exc)))


def _tokenize_worker(result_q):
    try:
        # A whitespace blob has no paragraph breaks, so tokenize() legitimately
        # raises ValueError after the (now linear) scan. Either outcome means the
        # call terminated rather than hanging.
        try:
            _tokenizer().tokenize(_CRAFTED_TEXT)
        except ValueError:
            pass
        result_q.put(("ok", "terminated"))
    except BaseException as exc:
        result_q.put(("error", repr(exc)))


def _run_in_process(target):
    """Run ``target(result_q)`` in a spawned process with a timeout.

    Returns ``(finished, status, payload)``. If the worker overruns ``_TIMEOUT``
    it is terminated (no lingering CPU) and ``finished`` is ``False``.
    """
    ctx = multiprocessing.get_context("spawn")
    result_q = ctx.Queue()
    proc = ctx.Process(target=target, args=(result_q,))
    proc.start()
    proc.join(_TIMEOUT)
    if proc.is_alive():
        proc.terminate()
        proc.join()
        return False, None, None
    try:
        status, payload = result_q.get_nowait()
    except queue.Empty:
        return True, "error", "worker produced no result"
    return True, status, payload


def test_mark_paragraph_breaks_preserves_behavior():
    """The possessive pattern must find the same paragraph breaks as before."""
    tt = _tokenizer()
    # a blank line between two paragraphs >= MIN_PARAGRAPH (100) apart is a break
    assert tt._mark_paragraph_breaks("x" * 120 + "\n\n" + "y" * 120) == [0, 120]
    # horizontal whitespace inside the break does not change the break position
    # (the match still starts where the trailing whitespace of para one begins)
    assert tt._mark_paragraph_breaks("x" * 120 + "  \n  \n  " + "y" * 120) == [0, 120]
    # no blank line -> only the implicit break at position 0
    assert tt._mark_paragraph_breaks("a single line with no breaks") == [0]
    # breaks closer than MIN_PARAGRAPH (100) are not recorded
    assert tt._mark_paragraph_breaks("x" * 50 + "\n\n" + "y" * 50) == [0]


def test_mark_paragraph_breaks_is_linear_on_whitespace_blob():
    """A long whitespace run with no blank line must not blow up (ReDoS)."""
    finished, status, payload = _run_in_process(_mark_worker)
    assert finished, "_mark_paragraph_breaks hung on a whitespace blob (ReDoS)"
    assert status == "ok", f"worker raised: {payload}"
    assert payload == [0]


def test_tokenize_does_not_hang_on_whitespace_blob():
    """End-to-end: tokenizing a whitespace blob must terminate (not hang)."""
    finished, status, payload = _run_in_process(_tokenize_worker)
    assert finished, "TextTilingTokenizer.tokenize hung on a whitespace blob (ReDoS)"
    assert status == "ok", f"tokenize raised in worker: {payload}"



"""Regression test for ReDoS in CCG lexicon parsing (CWE-1333).

``LEX_RE`` parses each lexicon line. Its identifier group ``[\\S_]+`` and the
arrow alternative ``[-=]+>`` both match ``-``/``=``, so a line with a long
``-``/``=`` run and no closing ``>`` made the engine slide the boundary across the
run and re-scan for the absent arrow, quadratically. Requiring whitespace before
the separator fixes the boundary, making parsing linear.

The "must not hang" tests run the work in a separate process (spawn) with a hard
timeout and ``terminate()`` on overrun, so a regression to the quadratic regex
cannot keep burning CPU for the rest of the suite.
"""

import multiprocessing
import queue

from nltk.ccg.lexicon import LEX_RE, fromstring

# One lexicon line: an identifier then a long run of '=' with no closing '>', so
# the arrow can never complete. ~tens of seconds with the old quadratic regex;
# sub-millisecond now (after process startup).
_CRAFTED = "a" + "=" * 200_000
_TIMEOUT = 15


def _regex_worker(result_q):
    try:
        result_q.put(("ok", LEX_RE.match(_CRAFTED) is not None))
    except BaseException as exc:  # surface to the parent process
        result_q.put(("error", repr(exc)))


def _fromstring_worker(result_q):
    try:
        # A line with no valid entry makes fromstring raise (AttributeError on
        # the None match); the point is that it must *terminate*, not hang.
        fromstring(_CRAFTED)
        result_q.put(("ok", "parsed"))
    except Exception:
        result_q.put(("ok", "raised"))
    except BaseException as exc:
        result_q.put(("error", repr(exc)))


def _run_in_process(target):
    ctx = multiprocessing.get_context("spawn")
    result_q = ctx.Queue()
    proc = ctx.Process(target=target, args=(result_q,))
    proc.start()
    proc.join(_TIMEOUT)
    if proc.is_alive():
        proc.terminate()
        proc.join()
        return False, None, None
    try:
        status, payload = result_q.get_nowait()
    except queue.Empty:
        return True, "error", "worker produced no result"
    return True, status, payload


def test_lex_re_does_not_hang():
    """LEX_RE must process a crafted lexicon line in linear time (no ReDoS)."""
    finished, status, value = _run_in_process(_regex_worker)
    assert finished, "LEX_RE hung on a crafted lexicon line (ReDoS)"
    assert status == "ok", f"worker raised: {value}"


def test_fromstring_does_not_hang():
    """End-to-end: fromstring() must terminate on a crafted lexicon line."""
    finished, status, value = _run_in_process(_fromstring_worker)
    assert finished, "fromstring() hung on a crafted lexicon line (ReDoS)"
    assert status == "ok", f"worker raised unexpectedly: {value}"


def test_lex_re_parses_spaced_and_compact_entries():
    """The ReDoS fix keeps both the whitespace-separated and the compact
    ``ident<sep>rhs`` forms working, and the identifier no longer absorbs part of
    the arrow (e.g. ``a-->b`` splits as ``a``/``-->``/``b``, not ``a-``/``->``)."""
    assert LEX_RE.match("the => Det").groups() == ("the", "=>", "Det")
    assert LEX_RE.match("Det :: NP/N").groups() == ("Det", "::", "NP/N")
    assert LEX_RE.match("a-->b").groups() == ("a", "-->", "b")
    assert LEX_RE.match("Det::NP/N").groups() == ("Det", "::", "NP/N")



"""Regression tests for nltk.translate.Alignment.

Covers the unbounded-allocation DoS (CWE-770; CVE-2026-12837): a tiny pair with a
huge left index used to force a dense per-left-index list allocation, because the
index was ``[[] for _ in range(self._len + 1)]``. The index is now sparse (a dict
keyed by the left indices that occur). These tests confirm the index is sparse and
that the public indexing / range behaviour is preserved.

The allocation test runs in a spawned process with a hard timeout so a regression
to the dense allocation cannot OOM or hang the rest of the suite.
"""

import multiprocessing
import queue

import pytest

from nltk.translate.api import Alignment


def test_index_is_sparse_not_dense():
    a = Alignment([(0, 0), (0, 1), (1, 2), (2, 2)])
    a.range()  # builds the index
    assert isinstance(a._index, dict)
    # keyed only by the left indices that occur, not range(_len + 1)
    assert set(a._index) == {0, 1, 2}


def test_getitem_and_range_preserved():
    a = Alignment([(0, 0), (0, 1), (1, 2), (2, 2)])
    assert sorted(a[0]) == [(0, 0), (0, 1)]
    assert sorted(a[2]) == [(2, 2)]
    assert a[5] == []  # a left index with no alignments
    assert a.range() == [0, 1, 2]
    assert a.range([0]) == [0, 1]
    assert a.range([1, 2]) == [2]
    assert sorted(a.invert()[2]) == [(2, 1), (2, 2)]


def test_getitem_rejects_non_integer_keys():
    # The sparse index has no contiguous range to slice; non-integer keys must
    # raise rather than silently return [] (which would mask caller bugs).
    a = Alignment([(0, 0), (0, 1), (1, 2)])
    for bad in (slice(0, 2), "0", 1.0, (0, 1)):
        with pytest.raises(TypeError):
            a[bad]


# A left index large enough to make a dense ``[[] for _ in range(_HUGE + 1)]``
# regression unmistakable (the index would hold _HUGE + 1 keys instead of 2),
# yet small enough that even if such a dense list were built it stays ~128 MB --
# detected by the key-count assertion below without risking an OOM of the host.
_HUGE = 2_000_000
_TIMEOUT = 15


def _alloc_worker(result_q):
    try:
        a = Alignment.fromstring(f"0-0 {_HUGE}-1")
        a.range()  # triggers _build_index
        _ = a[0]
        result_q.put(("ok", len(a._index)))
    except BaseException as exc:  # surface to the parent process
        result_q.put(("error", repr(exc)))


def _run_in_process(target):
    ctx = multiprocessing.get_context("spawn")
    result_q = ctx.Queue()
    proc = ctx.Process(target=target, args=(result_q,))
    proc.start()
    proc.join(_TIMEOUT)
    if proc.is_alive():
        proc.terminate()
        proc.join()
        return False, None, None
    try:
        status, payload = result_q.get_nowait()
    except queue.Empty:
        return True, "error", "worker produced no result"
    return True, status, payload


def test_huge_left_index_does_not_allocate():
    """A tiny pair with a huge left index must not allocate a dense index."""
    finished, status, value = _run_in_process(_alloc_worker)
    assert finished, "Alignment allocated a dense index for a huge left index (DoS)"
    assert status == "ok", f"worker raised: {value}"
    assert value == 2, f"index must be sparse (2 keys), got {value}"



import hashlib
import os
import shutil
import unittest
import unittest.mock
import xml.etree.ElementTree as ET

from nltk import download
from nltk.downloader import Downloader, Package, build_index


class TestPackageFromXmlInjection(unittest.TestCase):
    """Security tests for XML attribute injection via Package.fromxml."""

    def test_fromxml_neutralises_injected_filename(self):
        """Malicious filename attribute in XML must be ignored."""
        mock_xml = """
        <package id="test_pkg" name="Test Package" subdir="corpora"
                 url="http://example.com/test_pkg.zip" size="100"
                 unzipped_size="100" checksum="0" unzip="0"
                 filename="../../../vulnerable_overwrite.txt" />
        """
        xml_element = ET.fromstring(mock_xml)
        pkg = Package.fromxml(xml_element)

        expected = os.path.join("corpora", "test_pkg.zip")
        self.assertEqual(pkg.filename, expected)
        self.assertNotIn("..", pkg.filename)
        self.assertFalse(os.path.isabs(pkg.filename))


def test_downloader_using_existing_parent_download_dir(tmp_path):
    """Test that download works properly when the parent folder of the download_dir exists"""

    download_dir = str(tmp_path.joinpath("another_dir"))
    download_status = download("mwa_ppdb", download_dir)
    assert download_status is True


def test_downloader_using_non_existing_parent_download_dir(tmp_path):
    """Test that download works properly when the parent folder of the download_dir does not exist"""

    download_dir = str(
        tmp_path.joinpath("non-existing-parent-folder", "another-non-existing-folder")
    )
    download_status = download("mwa_ppdb", download_dir)
    assert download_status is True


def test_downloader_redownload(tmp_path):
    """Test that a second download correctly triggers the 'already up-to-date' message"""

    first_download = 0
    second_download = 1

    download_dir = str(tmp_path.joinpath("test_repeat_download"))
    for i in range(first_download, second_download + 1):
        # capsys doesn't capture functools.partial stdout, which nltk.download.show uses, so just mock print
        with unittest.mock.patch("builtins.print") as print_mock:
            download_status = download("stopwords", download_dir)
            assert download_status is True
            if i == first_download:
                expected_second_call = unittest.mock.call(
                    "[nltk_data]   Unzipping %s."
                    % os.path.join("corpora", "stopwords.zip")
                )
                assert print_mock.call_args_list[1].args == expected_second_call.args
            elif i == second_download:
                expected_second_call = unittest.mock.call(
                    "[nltk_data]   Package stopwords is already up-to-date!"
                )
                assert print_mock.call_args_list[1].args == expected_second_call.args


def test_build_index(tmp_path):
    """Test building index with both checksums."""

    test_pkg_dir = str(tmp_path.joinpath("packages"))
    test_pkg_name = "test_package"
    test_pkg_path = os.path.join(test_pkg_dir, f"{test_pkg_name}")
    os.makedirs(test_pkg_path, exist_ok=True)
    test_xml_path = os.path.join(test_pkg_path, f"{test_pkg_name}.xml")
    with open(test_xml_path, "w") as fi:
        fi.write(
            f'<package id="{test_pkg_name}" name="A Test Package" webpage="http://www.somefake.url/"'
            ' unzip="1"/>'
        )
    # Cannot mock a zip here as we are trying to validate file checksums, so just create a simple one with the XML
    zip_path = os.path.join(test_pkg_path, f"{test_pkg_name}")
    shutil.make_archive(
        base_name=zip_path,
        format="zip",
        root_dir=test_pkg_dir,
        base_dir=os.path.basename(test_pkg_path),
    )
    xml_index = build_index(
        root=os.path.dirname(test_pkg_dir), base_url="https://someurl"
    )
    package_element = xml_index[0][0]
    assert package_element.get("id") == "test_package"
    md5_checksum = package_element.get("checksum")
    assert isinstance(md5_checksum, str)
    assert len(md5_checksum) > 5
    sha256_checksum = package_element.get("sha256_checksum")
    assert isinstance(sha256_checksum, str)
    assert len(sha256_checksum) > 5


def test_download_package_uses_scoped_pathsec_open(tmp_path):
    """
    Regression test for PR #3622.

    Verify that _download_package() routes the package write through
    pathsec.open with the downloader's download_dir as required_root, and that
    the final destination is still validated before os.replace().
    """
    download_dir = str(tmp_path.joinpath("download"))
    os.makedirs(download_dir, exist_ok=True)

    downloader = Downloader(download_dir=download_dir)

    class DummyInfo:
        id = "dummy"
        url = "https://example.com/dummy.txt"
        size = 1
        filename = os.path.join("corpora", "dummy.txt")
        subdir = "corpora"
        unzip = False
        sha256_checksum = hashlib.sha256(b"x").hexdigest()
        checksum = hashlib.md5(b"x").hexdigest()

    info = DummyInfo()
    tmp_file = os.path.join(download_dir, info.filename) + ".tmp"
    final_file = os.path.join(download_dir, info.filename)

    with unittest.mock.patch(
        "nltk.downloader.urlopen"
    ) as urlopen_mock, unittest.mock.patch(
        "nltk.downloader.pathsec_open"
    ) as pathsec_open_mock, unittest.mock.patch(
        "nltk.downloader.validate_path"
    ) as validate_path_mock, unittest.mock.patch(
        "nltk.downloader.os.replace"
    ) as replace_mock, unittest.mock.patch(
        "nltk.downloader.os.makedirs"
    ), unittest.mock.patch(
        "nltk.downloader.os.open", return_value=3
    ), unittest.mock.patch(
        "nltk.downloader.os.close"
    ), unittest.mock.patch(
        "nltk.downloader.os.path.exists", return_value=False
    ), unittest.mock.patch(
        "nltk.downloader.time.time", return_value=0
    ), unittest.mock.patch(
        "nltk.downloader.itertools.count", return_value=iter([0])
    ), unittest.mock.patch(
        "nltk.downloader.os.path.getsize", return_value=1
    ), unittest.mock.patch(
        "nltk.downloader.sha256_hexdigest",
        return_value=hashlib.sha256(b"x").hexdigest(),
    ), unittest.mock.patch(
        "nltk.downloader.md5_hexdigest",
        return_value=hashlib.md5(b"x").hexdigest(),
    ), unittest.mock.patch.object(
        Downloader, "status", return_value=Downloader.NOT_INSTALLED
    ):

        infile = unittest.mock.Mock()
        infile.read.side_effect = [b"x", b""]
        infile.close.return_value = None
        urlopen_mock.return_value = infile

        outfile = unittest.mock.Mock()
        pathsec_open_mock.return_value.__enter__.return_value = outfile
        pathsec_open_mock.return_value.__exit__.return_value = False

        list(downloader._download_package(info, download_dir, force=True))

        pathsec_open_mock.assert_called_once_with(
            tmp_file,
            "wb",
            context="Downloader._download_package",
            required_root=download_dir,
        )

        validate_path_mock.assert_any_call(
            final_file,
            context="Downloader._download_package",
            required_root=download_dir,
        )

        replace_mock.assert_called_once_with(tmp_file, final_file)

        # Sanity check: the PR-specific path scoping does not change builtins.open.
        assert open is __builtins__["open"]



# Natural Language Toolkit: Interface to the CRFSuite Tagger
#
# Copyright (C) 2001-2026 NLTK Project
# Author: Long Duong <longdt219@gmail.com>
#         John Winstead <https://github.com/jhnwnstd> (fixes)
# URL: <https://www.nltk.org/>
# For license information, see LICENSE.TXT

"""
A module for POS tagging using CRFSuite
"""

import re
import unicodedata
import warnings

from nltk.tag.api import TaggerI

try:
    import pycrfsuite
except ImportError:
    pycrfsuite = None

# Punctuation categories from the Unicode general-category table.
# Module-level so the per-token feature loop doesn't rebuild it.
_PUNC_CATEGORIES = frozenset({"Pc", "Pd", "Ps", "Pe", "Pi", "Pf", "Po"})


class CRFTagger(TaggerI):
    """
    A module for POS tagging using CRFSuite https://pypi.python.org/pypi/python-crfsuite

    >>> from nltk.tag import CRFTagger
    >>> ct = CRFTagger()  # doctest: +SKIP

    >>> train_data = [
    ...     [('University','Noun'), ('is','Verb'), ('a','Det'),
    ...      ('good','Adj'), ('place','Noun')],
    ...     [('dog','Noun'), ('eat','Verb'), ('meat','Noun')],
    ... ]

    >>> ct.train(train_data, 'model.crf.tagger')  # doctest: +SKIP
    >>> ct.tag_sents([['dog','is','good'], ['Cat','eat','meat']])  # doctest: +SKIP
    [[('dog', 'Noun'), ('is', 'Verb'), ('good', 'Adj')],
     [('Cat', 'Noun'), ('eat', 'Verb'), ('meat', 'Noun')]]

    >>> gold_sentences = [
    ...     [('dog','Noun'), ('is','Verb'), ('good','Adj')],
    ...     [('Cat','Noun'), ('eat','Verb'), ('meat','Noun')],
    ... ]
    >>> ct.accuracy(gold_sentences)  # doctest: +SKIP
    1.0

    Setting learned model file
    >>> ct = CRFTagger()  # doctest: +SKIP
    >>> ct.set_model_file('model.crf.tagger')  # doctest: +SKIP
    >>> ct.accuracy(gold_sentences)  # doctest: +SKIP
    1.0

    Default features are cached by token surface form. Call
    ``clear_feature_cache()`` to drop the cache (e.g. between long-running
    open-vocabulary tagging passes). A custom ``feature_func`` bypasses
    the cache.
    """

    def __init__(self, feature_func=None, verbose=False, training_opt=None):
        """
        Initialize the CRFSuite tagger

        :param feature_func: Function that extracts features for each token
            of a sentence. Takes two parameters (``tokens``, ``idx``) and
            returns the feature list for ``tokens[idx]``. See the built-in
            ``_get_features`` for an example.
        :param verbose: Emit debugging messages during training.
        :type verbose: bool
        :param training_opt: python-crfsuite training options.
        :type training_opt: dict

        Set of possible training options (using LBFGS training algorithm).
            :'feature.minfreq': Minimum frequency of features.
            :'feature.possible_states': Force generating possible state features.
            :'feature.possible_transitions': Force generating possible transition
                features.
            :'c1': Coefficient for L1 regularization.
            :'c2': Coefficient for L2 regularization.
            :'max_iterations': Maximum number of iterations for L-BFGS.
            :'num_memories': Number of limited memories for approximating the
                inverse Hessian matrix.
            :'epsilon': Epsilon for testing the convergence of the objective.
            :'period': Duration of iterations for the stopping criterion.
            :'delta': Threshold for the stopping criterion; L-BFGS stops when
                the log-likelihood improvement over the last ${period}
                iterations is no greater than this threshold.
            :'linesearch': Line search algorithm used in L-BFGS updates:

                - 'MoreThuente': More and Thuente's method,
                - 'Backtracking': Backtracking method with regular Wolfe condition,
                - 'StrongBacktracking': Backtracking method with strong Wolfe condition
            :'max_linesearch':  Maximum number of trials for the line search.
        """

        if pycrfsuite is None:
            raise ImportError("CRFTagger requires python-crfsuite to be installed.")

        self._model_file = ""
        self._tagger = pycrfsuite.Tagger()

        if feature_func is None:
            self._feature_func = self._get_features
        else:
            self._feature_func = feature_func

        self._verbose = verbose
        # Avoid mutable default; copy so caller mutations don't leak in.
        self._training_options = {} if training_opt is None else dict(training_opt)
        self._pattern = re.compile(r"\d")
        # Avoid the module-level ``re.search`` dispatch in the feature loop.
        self._pattern_search = self._pattern.search
        # Token-keyed cache; default features are token-local. A custom
        # ``feature_func`` replaces ``_get_features`` and skips this dict.
        self._feature_cache = {}

    def set_model_file(self, model_file):
        self._model_file = model_file
        self._tagger.open(self._model_file)

    def _get_features(self, tokens, idx):
        """
        Extract basic features about this word including
            - Current word
            - is it capitalized?
            - Does it have punctuation?
            - Does it have a number?
            - Suffixes up to length 3

        Note that : we might include features over previous word, next word etc.

        :return: a list which contains the features
        :rtype: list(str)
        """
        token = tokens[idx]

        # Return a fresh list so callers can't mutate the cached tuple.
        cached = self._feature_cache.get(token)
        if cached is not None:
            return list(cached)

        feature_list = []

        if not token:
            self._feature_cache[token] = ()
            return feature_list

        append = feature_list.append
        token_len = len(token)

        if token[0].isupper():
            append("CAPITALIZATION")

        if self._pattern_search(token) is not None:
            append("HAS_NUM")

        category = unicodedata.category
        if all(category(ch) in _PUNC_CATEGORIES for ch in token):
            append("PUNCTUATION")

        if token_len > 1:
            append("SUF_" + token[-1:])
        if token_len > 2:
            append("SUF_" + token[-2:])
        if token_len > 3:
            append("SUF_" + token[-3:])

        append("WORD_" + token)

        self._feature_cache[token] = tuple(feature_list)
        return feature_list

    def clear_feature_cache(self):
        """Drop the default-feature cache.

        The cache is vocabulary-bounded, but long-running processes that
        ingest open-vocabulary streams (URLs, tweets, generated text) may
        want to reset it explicitly. A custom ``feature_func`` bypasses
        the cache, so this is a no-op for non-default extractors.
        """
        self._feature_cache.clear()

    def tag_sents(self, sentences=None, **kwargs):
        """
        Tag a list of sentences. Before using this function, the model file
        must be specified by either:

        - Train a new model using ``train`` function
        - Use the pre-trained model which is set via ``set_model_file`` function

        :param sentences: sentences to tag.
        :type sentences: list(list(str))
        :param sents: deprecated alias for ``sentences``; emits ``DeprecationWarning``.
        :return: list of tagged sentences.
        :rtype: list(list(tuple(str,str)))
        """
        if "sents" in kwargs:
            if sentences is not None:
                raise TypeError(
                    "tag_sents() got both 'sentences' and 'sents'; "
                    "use 'sentences' only ('sents' is deprecated)."
                )
            warnings.warn(
                "tag_sents(sents=...) is deprecated; use tag_sents(sentences=...).",
                DeprecationWarning,
                stacklevel=2,
            )
            sentences = kwargs.pop("sents")
        if kwargs:
            unexpected = ", ".join(sorted(kwargs))
            raise TypeError(
                f"tag_sents() got unexpected keyword arguments: {unexpected}"
            )
        if sentences is None:
            raise TypeError("tag_sents() missing 1 required argument: 'sentences'")

        if isinstance(sentences, (str, bytes)):
            raise TypeError(
                "tag_sents() expects a list of tokenized sentences, "
                "not a single string or bytes object."
            )

        # Catch ``[token, token, ...]`` (one tokenized sentence) before the
        # model-file check so the error is clear even on an untrained tagger.
        # Generators fall through to the per-iteration check below.
        if (
            isinstance(sentences, (list, tuple))
            and sentences
            and isinstance(sentences[0], (str, bytes))
        ):
            raise TypeError(
                "tag_sents() expects a list of tokenized sentences, "
                "not a single tokenized sentence."
            )

        if self._model_file == "":
            raise RuntimeError(
                "No model file set; call train() or set_model_file() first."
            )

        # Hoist hot-loop attribute lookups out of the per-sentence loop.
        feature_func = self._feature_func
        tag = self._tagger.tag

        result = []
        for tokens in sentences:
            if isinstance(tokens, (str, bytes)):
                # Safety net for generator inputs the up-front check skipped.
                raise TypeError(
                    "tag_sents() expects a list of tokenized sentences, "
                    "not a single tokenized sentence."
                )
            features = [feature_func(tokens, i) for i in range(len(tokens))]
            labels = tag(features)

            if len(labels) != len(tokens):
                raise RuntimeError(
                    f"CRF returned {len(labels)} labels for {len(tokens)} tokens."
                )

            tagged_sent = list(zip(tokens, labels, strict=True))
            result.append(tagged_sent)

        return result

    def train(self, train_data, model_file):
        """
        Train the CRF tagger using CRFSuite

        :param train_data: list of annotated sentences.
        :type train_data: list(list(tuple(str,str)))
        :param model_file: path where the trained model will be written.
        :type model_file: str
        """
        if pycrfsuite is None:
            raise ImportError("CRFTagger requires python-crfsuite to be installed.")

        trainer = pycrfsuite.Trainer(verbose=self._verbose)
        trainer.set_params(self._training_options)

        feature_func = self._feature_func
        append = trainer.append

        for sent in train_data:
            tokens, labels = zip(*sent, strict=True)
            features = [feature_func(tokens, i) for i in range(len(tokens))]
            append(features, labels)

        trainer.train(model_file)
        self.set_model_file(model_file)

    def tag(self, tokens):
        """
        Tag a sentence using the python-crfsuite tagger. Before using this
        function, the model file must be specified by either:

        - Train a new model using ``train`` function
        - Use the pre-trained model which is set via ``set_model_file`` function

        :param tokens: tokens to tag.
        :type tokens: list(str)
        :return: list of tagged tokens.
        :rtype: list(tuple(str,str))
        """
        if isinstance(tokens, (str, bytes)):
            raise TypeError(
                "tag() expects a list of tokens, not a single string or bytes. "
                "Tokenize first (e.g. tokens.split() or word_tokenize(s))."
            )
        return self.tag_sents([tokens])[0]



import unittest

from nltk.corpus import ieer


class TestIEER(unittest.TestCase):
    def test_parsed_docs(self):
        docs = ieer.parsed_docs("APW_19980314")
        self.assertEqual(len(docs), 23)
        self.assertEqual(docs[0].docno, "APW19980314.0391")

    def test_docs_does_not_contain_empty_strings(self):
        docs = list(ieer.docs())
        # The corpus has 94 valid documents in total
        self.assertEqual(len(docs), 94)
        # Ensure there are no trailing empty documents/strings resulting from EOF misbehavior
        self.assertTrue(all(bool(doc.strip()) for doc in docs))



"""Regression tests for the untrusted-search-path fix in MaltParser (CWE-426).

``find_maltparser`` used ``os.path.exists(parser_dirname)`` to accept a directory
as-is. For a *relative* ``parser_dirname`` (as NLTK's docs use) that resolves
against the current working directory and is checked *before* the ``MALT_PARSER``
environment variable, so an attacker who can write a ``maltparser-*/`` directory
into the CWD could place its jars on the Java classpath -- overriding a trusted
``MALT_PARSER`` -- and have ``org.maltparser.Malt`` executed (RCE).

Only an explicit *absolute* directory is now taken as-is; a relative name is
resolved through ``MALT_PARSER``.
"""

import os
import pathlib

import pytest

from nltk.parse.malt import find_maltparser

# The names find_maltparser asserts must be present in the directory.
_MALT_JARS = ["log4j.jar", "libsvm.jar", "liblinear-1.8.jar", "maltparser-1.9.2.jar"]


def _make_malt_dir(path):
    path.mkdir(parents=True, exist_ok=True)
    for jar in _MALT_JARS:
        (path / jar).write_bytes(b"")
    return path


def test_relative_dirname_not_resolved_against_cwd(tmp_path, monkeypatch):
    """A relative parser_dirname matching a CWD directory must NOT be used."""
    cwd = tmp_path / "cwd"
    cwd.mkdir()
    _make_malt_dir(cwd / "maltparser-1.9.2")  # attacker-planted in the CWD
    monkeypatch.chdir(cwd)
    monkeypatch.delenv("MALT_PARSER", raising=False)

    # With no MALT_PARSER configured, the relative name must not fall back to the
    # current working directory; discovery fails instead of silently using it.
    with pytest.raises(LookupError):
        find_maltparser("maltparser-1.9.2")


def test_cwd_does_not_override_configured_malt_parser(tmp_path, monkeypatch):
    """A CWD directory must not shadow a configured MALT_PARSER."""
    cwd = tmp_path / "cwd"
    cwd.mkdir()
    _make_malt_dir(cwd / "maltparser-1.9.2")  # attacker-planted in the CWD
    trusted = _make_malt_dir(tmp_path / "trusted")  # the configured location
    monkeypatch.chdir(cwd)
    monkeypatch.setenv("MALT_PARSER", str(trusted))

    jars = find_maltparser("maltparser-1.9.2")
    chosen = os.path.realpath(os.path.dirname(jars[0]))
    assert chosen == os.path.realpath(
        str(trusted)
    ), "CWD directory overrode the trusted MALT_PARSER"


def test_absolute_path_still_accepted(tmp_path, monkeypatch):
    """An explicit absolute directory is still used as-is."""
    abs_dir = _make_malt_dir(tmp_path / "absmalt")
    monkeypatch.delenv("MALT_PARSER", raising=False)

    jars = find_maltparser(str(abs_dir))
    chosen = os.path.realpath(os.path.dirname(jars[0]))
    assert chosen == os.path.realpath(str(abs_dir))


def test_pathlike_argument_is_handled(tmp_path, monkeypatch):
    """A pathlib.Path argument is accepted; a relative one fails cleanly."""
    abs_dir = _make_malt_dir(tmp_path / "absmalt")
    monkeypatch.delenv("MALT_PARSER", raising=False)
    # An absolute Path is accepted (normalised via os.fspath).
    assert find_maltparser(pathlib.Path(abs_dir))
    # A relative Path with no MALT_PARSER raises a clean LookupError -- not an
    # AssertionError from find_dir()'s str check, and not a CWD pickup.
    monkeypatch.chdir(tmp_path)
    with pytest.raises(LookupError):
        find_maltparser(pathlib.Path("not-a-configured-malt-dir"))
